import numpy as np
import copy
from lenstronomy.PointSource.point_source_cached import PointSourceCached

__all__ = ["PointSource"]

_SUPPORTED_MODELS = ["UNLENSED", "LENSED_POSITION", "SOURCE_POSITION"]


class PointSource(object):
    def __init__(
        self,
        point_source_type_list,
        lensModel=None,
        fixed_magnification_list=None,
        additional_images_list=None,
        flux_from_point_source_list=None,
        magnification_limit=None,
        save_cache=False,
        kwargs_lens_eqn_solver=None,
        index_lens_model_list=None,
        point_source_frame_list=None,
    ):
        """

        :param point_source_type_list: list of point source types
        :param lensModel: instance of the LensModel() class
        :param fixed_magnification_list: list of booleans (same length as point_source_type_list).
         If True, magnification ratio of point sources is fixed to the one given by the lens model.
         This option then requires to provide a 'source_amp' amplitude of the source brightness instead of
         'point_amp' the list of image brightnesses.
        :param additional_images_list: list of booleans (same length as point_source_type_list). If True, search for
         additional images of the same source is conducted.
        :param flux_from_point_source_list: list of booleans (optional), if set, will only return image positions
         (for imaging modeling) for the subset of the point source lists that =True. This option enables to model
         imaging data with transient point sources, when the point source positions are measured and present at a
         different time than the imaging data
        :param magnification_limit: float >0 or None, if float is set and additional images are computed, only those
         images will be computed that exceed the lensing magnification (absolute value) limit
        :param save_cache: bool, saves image positions and only if delete_cache is executed, a new solution of the lens
         equation is conducted with the lens model parameters provided. This can increase the speed as multiple times
         the image positions are requested for the same lens model. Attention in usage!
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical settings for the lens equation solver
         see LensEquationSolver() class for details, such as:
         min_distance=0.01, search_window=5, precision_limit=10**(-10), num_iter_max=100
        :param index_lens_model_list: list (length of different patches/bands) of integer lists, e.g., [[0, 1], [2, 3]];
         evaluating a subset of the lens models per individual bands. If this keyword is set, the image positions need
         to have a specified band/frame assigned to it
        :param point_source_frame_list: list of lists mirroring the structure of the image positions.
         Integers correspond to the i'th list entry of index_lens_model_list indicating in which frame/band the image is
         appearing, e.g., four images and four cutouts, you can do [[0], [1], [2], [3]] when the frames
         (ordered as the list) are in the same order as the point sources (indices)
        """
        if len(point_source_type_list) > 0:
            if index_lens_model_list is not None and point_source_frame_list is None:
                raise ValueError(
                    "with specified index_lens_model_list a specified point_source_frame_list argument is "
                    "required"
                )
            if index_lens_model_list is None:
                point_source_frame_list = [None] * len(point_source_type_list)
        self._index_lens_model_list = index_lens_model_list
        self._point_source_frame_list = point_source_frame_list
        self._lensModel = lensModel
        self.point_source_type_list = point_source_type_list
        self._point_source_list = []
        if fixed_magnification_list is None:
            fixed_magnification_list = [False] * len(point_source_type_list)
        self._fixed_magnification_list = fixed_magnification_list
        if additional_images_list is None:
            additional_images_list = [False] * len(point_source_type_list)
        if flux_from_point_source_list is None:
            flux_from_point_source_list = [True] * len(point_source_type_list)
        self._flux_from_point_source_list = flux_from_point_source_list
        for i, model in enumerate(point_source_type_list):
            if model == "UNLENSED":
                from lenstronomy.PointSource.Types.unlensed import Unlensed

                self._point_source_list.append(
                    PointSourceCached(Unlensed(), save_cache=save_cache)
                )
            elif model == "LENSED_POSITION":
                from lenstronomy.PointSource.Types.lensed_position import (
                    LensedPositions,
                )

                self._point_source_list.append(
                    PointSourceCached(
                        LensedPositions(
                            lensModel,
                            fixed_magnification=fixed_magnification_list[i],
                            additional_images=additional_images_list[i],
                            index_lens_model_list=index_lens_model_list,
                            point_source_frame_list=point_source_frame_list[i],
                        ),
                        save_cache=save_cache,
                    )
                )
            elif model == "SOURCE_POSITION":
                from lenstronomy.PointSource.Types.source_position import (
                    SourcePositions,
                )

                self._point_source_list.append(
                    PointSourceCached(
                        SourcePositions(
                            lensModel, fixed_magnification=fixed_magnification_list[i]
                        ),
                        save_cache=save_cache,
                    )
                )
            else:
                raise ValueError(
                    "Point-source model %s not available. Supported models are %s ."
                    % (model, _SUPPORTED_MODELS)
                )
        if kwargs_lens_eqn_solver is None:
            kwargs_lens_eqn_solver = {}
        self._kwargs_lens_eqn_solver = kwargs_lens_eqn_solver
        self._magnification_limit = magnification_limit
        self._save_cache = save_cache

    def update_search_window(
        self,
        search_window,
        x_center,
        y_center,
        min_distance=None,
        only_from_unspecified=False,
    ):
        """Update the search area for the lens equation solver.

        :param search_window: search_window: window size of the image position search
            with the lens equation solver.
        :param x_center: center of search window
        :param y_center: center of search window
        :param min_distance: minimum search distance
        :param only_from_unspecified: bool, if True, only sets keywords that previously
            have not been set
        :return: updated self instances
        """
        if (
            min_distance is not None
            and "min_distance" not in self._kwargs_lens_eqn_solver
            and only_from_unspecified
        ):
            self._kwargs_lens_eqn_solver["min_distance"] = min_distance
        if only_from_unspecified:
            self._kwargs_lens_eqn_solver[
                "search_window"
            ] = self._kwargs_lens_eqn_solver.get("search_window", search_window)
            self._kwargs_lens_eqn_solver["x_center"] = self._kwargs_lens_eqn_solver.get(
                "x_center", x_center
            )
            self._kwargs_lens_eqn_solver["y_center"] = self._kwargs_lens_eqn_solver.get(
                "y_center", y_center
            )
        else:
            self._kwargs_lens_eqn_solver["search_window"] = search_window
            self._kwargs_lens_eqn_solver["x_center"] = x_center
            self._kwargs_lens_eqn_solver["y_center"] = y_center

    def update_lens_model(self, lens_model_class):
        """

        :param lens_model_class: instance of LensModel class
        :return: update instance of lens model class
        """
        self.delete_lens_model_cache()
        self._lensModel = lens_model_class
        for model in self._point_source_list:
            model.update_lens_model(lens_model_class=lens_model_class)

    def delete_lens_model_cache(self):
        """Deletes the variables saved for a specific lens model.

        :return: None
        """
        for model in self._point_source_list:
            model.delete_lens_model_cache()

    def set_save_cache(self, save_cache):
        """Set the save cache boolean to new value.

        :param save_cache: bool, if True, saves (or uses a previously saved) values
        :return: updated class and sub-class instances to either save or not save the
            point source information in cache
        """
        self._set_save_cache(save_cache)
        self._save_cache = save_cache

    def _set_save_cache(self, save_cache):
        """Set the save cache boolean to new value. This function is for use within this
        class for temporarily set the cache within a single routine.

        :param save_cache: bool, if True, saves (or uses a previously saved) values
        :return: None
        """
        for model in self._point_source_list:
            model.set_save_cache(save_cache)

    def k_list(self, k):
        """

        :param k: index of point source model
        :return: list of lengths of images with corresponding lens models in the frame (or None if not multi-frame)
        """
        if self._index_lens_model_list is not None:
            k_list = []
            for point_source_frame in self._point_source_frame_list[k]:
                k_list.append(self._index_lens_model_list[point_source_frame])
        else:
            k_list = None
        return k_list

    def source_position(self, kwargs_ps, kwargs_lens):
        """Intrinsic source positions of the point sources.

        :param kwargs_ps: keyword argument list of point source models
        :param kwargs_lens: keyword argument list of lens models
        :return: list of source positions for each point source model
        """
        x_source_list = []
        y_source_list = []
        for i, model in enumerate(self._point_source_list):
            kwargs = kwargs_ps[i]
            x_source, y_source = model.source_position(kwargs, kwargs_lens)
            x_source_list.append(x_source)
            y_source_list.append(y_source)
        return x_source_list, y_source_list

    def image_position(
        self,
        kwargs_ps,
        kwargs_lens,
        k=None,
        original_position=False,
        additional_images=False,
    ):
        """Image positions as observed on the sky of the point sources.

        :param kwargs_ps: point source parameter keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param k: None, int or boolean list; only returns a subset of the model
            predictions
        :param original_position: boolean (only applies to 'LENSED_POSITION' models),
            returns the image positions in the model parameters and does not re-compute
            images (which might be differently ordered) in case of the lens equation
            solver
        :param additional_images: if True, solves the lens equation for additional
            images
        :type additional_images: bool
        :return: list of: list of image positions per point source model component
        """
        x_image_list = []
        y_image_list = []
        for i, model in enumerate(self._point_source_list):
            if k is None or k == i:
                kwargs = kwargs_ps[i]
                x_image, y_image = model.image_position(
                    kwargs,
                    kwargs_lens,
                    magnification_limit=self._magnification_limit,
                    kwargs_lens_eqn_solver=self._kwargs_lens_eqn_solver,
                    additional_images=additional_images,
                )
                # this takes action when new images are computed not necessary in order
                if (
                    original_position is True
                    and additional_images is True
                    and self.point_source_type_list[i] == "LENSED_POSITION"
                ):
                    x_o, y_o = kwargs["ra_image"], kwargs["dec_image"]
                    x_image, y_image = _sort_position_by_original(
                        x_o, y_o, x_image, y_image
                    )

                x_image_list.append(x_image)
                y_image_list.append(y_image)
        return x_image_list, y_image_list

    def point_source_list(self, kwargs_ps, kwargs_lens, k=None, with_amp=True):
        """Returns the coordinates and amplitudes of all point sources in a single
        array.

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param k: None, int or list of int's to select a subset of the point source
            models in the return
        :param with_amp: bool, if False, ignores the amplitude parameters in the return
            and instead provides ones for each point source image
        :return: ra_array, dec_array, amp_array
        """
        # here we save the cache of the individual models but do not overwrite the class boolean variable to do so
        self._set_save_cache(True)
        # we make sure we do not re-compute the image positions twice when evaluating position and their amplitudes
        ra_list, dec_list = self.image_position(kwargs_ps, kwargs_lens, k=k)
        if with_amp is True:
            amp_list = self.image_amplitude(kwargs_ps, kwargs_lens, k=k)
        else:
            amp_list = np.ones_like(ra_list)

        # here we delete the individual modeling caches in case this was the option
        if self._save_cache is False:
            self.delete_lens_model_cache()
            self._set_save_cache(self._save_cache)

        ra_array, dec_array, amp_array = [], [], []
        for i, ra in enumerate(ra_list):
            for j in range(len(ra)):
                ra_array.append(ra_list[i][j])
                dec_array.append(dec_list[i][j])
                amp_array.append(amp_list[i][j])
        return ra_array, dec_array, amp_array

    def num_basis(self, kwargs_ps, kwargs_lens):
        """Number of basis functions for linear inversion.

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :return: int
        """
        n = 0
        ra_pos_list, dec_pos_list = self.image_position(kwargs_ps, kwargs_lens)
        for i, model in enumerate(self.point_source_type_list):
            if self._flux_from_point_source_list[i]:
                if self._fixed_magnification_list[i]:
                    n += 1
                else:
                    n += len(ra_pos_list[i])
        return n

    def image_amplitude(self, kwargs_ps, kwargs_lens, k=None):
        """Returns the image amplitudes.

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param k: None, int or list of int's to select a subset of the point source
            models in the return
        :return: list of image amplitudes per model component
        """
        amp_list = []
        for i, model in enumerate(self._point_source_list):
            if (k is None or k == i) and self._flux_from_point_source_list[i]:
                amp_list.append(
                    model.image_amplitude(
                        kwargs_ps=kwargs_ps[i],
                        kwargs_lens=kwargs_lens,
                        kwargs_lens_eqn_solver=self._kwargs_lens_eqn_solver,
                    )
                )
        return amp_list

    def source_amplitude(self, kwargs_ps, kwargs_lens):
        """Intrinsic (unlensed) point source amplitudes.

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :return: list of intrinsic (unlensed) point source amplitudes
        """
        amp_list = []
        for i, model in enumerate(self._point_source_list):
            if self._flux_from_point_source_list[i]:
                amp_list.append(
                    model.source_amplitude(
                        kwargs_ps=kwargs_ps[i], kwargs_lens=kwargs_lens
                    )
                )
        return amp_list

    def linear_response_set(self, kwargs_ps, kwargs_lens=None, with_amp=False):
        """

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param with_amp: bool, if True returns the image amplitude derived from kwargs_ps,
         otherwise the magnification of the lens model
        :return: ra_pos, dec_pos, amp, n
        """
        ra_pos = []
        dec_pos = []
        amp = []
        self._set_save_cache(True)
        x_image_list, y_image_list = self.image_position(kwargs_ps, kwargs_lens)
        for i, model in enumerate(self._point_source_list):
            if self._flux_from_point_source_list[i]:
                x_pos = x_image_list[i]
                y_pos = y_image_list[i]
                if self._fixed_magnification_list[i]:
                    ra_pos.append(list(x_pos))
                    dec_pos.append(list(y_pos))
                    if with_amp:
                        mag = self.image_amplitude(kwargs_ps, kwargs_lens, k=i)[0]
                    else:
                        mag = self._lensModel.magnification(x_pos, y_pos, kwargs_lens)
                        mag = np.abs(mag)
                    amp.append(list(mag))
                else:
                    if with_amp:
                        mag = self.image_amplitude(kwargs_ps, kwargs_lens, k=i)[0]
                    else:
                        mag = np.ones_like(x_pos)
                    for j in range(len(x_pos)):
                        ra_pos.append([x_pos[j]])
                        dec_pos.append([y_pos[j]])
                        amp.append([mag[j]])
        n = len(ra_pos)
        if self._save_cache is False:
            self.delete_lens_model_cache()
            self._set_save_cache(self._save_cache)
        return ra_pos, dec_pos, amp, n

    def update_linear(self, param, i, kwargs_ps, kwargs_lens):
        """

        :param param: list of floats corresponding ot the parameters being sampled
        :param i: index of the first parameter relevant for this class
        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :return: kwargs_ps with updated linear parameters, index of the next parameter relevant for another class
        """
        ra_pos_list, dec_pos_list = self.image_position(kwargs_ps, kwargs_lens)
        for k, model in enumerate(self._point_source_list):
            if self._flux_from_point_source_list[k]:
                kwargs = kwargs_ps[k]
                if self._fixed_magnification_list[k]:
                    kwargs["source_amp"] = param[i]
                    i += 1
                else:
                    n_points = len(ra_pos_list[k])
                    kwargs["point_amp"] = np.array(param[i : i + n_points])
                    i += n_points
        return kwargs_ps, i

    def linear_param_from_kwargs(self, kwargs_list):
        """Inverse function of update_linear() returning the linear amplitude list for
        the keyword argument list.

        :param kwargs_list: model parameters including the linear amplitude parameters
        :type kwargs_list: list of keyword arguments
        :return: list of linear amplitude parameters
        :rtype: list
        """
        param = []
        for k, model in enumerate(self._point_source_list):
            if self._flux_from_point_source_list[k]:
                kwargs = kwargs_list[k]
                if self._fixed_magnification_list[k]:
                    param.append(kwargs["source_amp"])
                else:
                    for a in kwargs["point_amp"]:
                        param.append(a)
        return param

    def check_image_positions(self, kwargs_ps, kwargs_lens, tolerance=0.001):
        """Checks whether the point sources in kwargs_ps satisfy the lens equation with
        a tolerance (computed by ray-tracing in the source plane)

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param tolerance: Euclidian distance between the source positions ray-traced
            backwards to be tolerated
        :return: bool: True, if requirement on tolerance is fulfilled, False if not.
        """
        x_image_list, y_image_list = self.image_position(kwargs_ps, kwargs_lens)
        for i, model in enumerate(self.point_source_type_list):
            if model in ["LENSED_POSITION", "SOURCE_POSITION"]:
                x_pos = x_image_list[i]
                y_pos = y_image_list[i]
                x_source, y_source = self._lensModel.ray_shooting(
                    x_pos, y_pos, kwargs_lens
                )
                dist = np.sqrt(
                    (x_source - x_source[0]) ** 2 + (y_source - y_source[0]) ** 2
                )
                if np.max(dist) > tolerance:
                    return False
        return True

    def set_amplitudes(self, amp_list, kwargs_ps):
        """Translates the amplitude parameters into the convention of the keyword
        argument list currently only used in SimAPI to transform magnitudes to
        amplitudes in the lenstronomy conventions.

        :param amp_list: list of model amplitudes for each point source model
        :param kwargs_ps: list of point source keywords
        :return: overwrites kwargs_ps with new amplitudes
        """
        kwargs_list = copy.deepcopy(kwargs_ps)
        for i, model in enumerate(self.point_source_type_list):
            if self._flux_from_point_source_list[i]:
                amp = amp_list[i]
                if model == "UNLENSED":
                    kwargs_list[i]["point_amp"] = amp
                elif model in ["LENSED_POSITION", "SOURCE_POSITION"]:
                    if self._fixed_magnification_list[i] is True:
                        kwargs_list[i]["source_amp"] = amp
                    else:
                        kwargs_list[i]["point_amp"] = amp
        return kwargs_list

    @classmethod
    def check_positive_flux(cls, kwargs_ps):
        """Check whether inferred linear parameters are positive.

        :param kwargs_ps: point source keyword argument list
        :return: bool, True, if all 'point_amp' parameters are positive semi-definite
        """
        pos_bool = True
        for kwargs in kwargs_ps:
            if "point_amp" in kwargs:
                point_amp = kwargs["point_amp"]
                if not np.all(point_amp >= 0):
                    pos_bool = False
                    break
            if "source_amp" in kwargs:
                point_amp = kwargs["source_amp"]
                if not np.all(point_amp >= 0):
                    pos_bool = False
                    break
        return pos_bool


def _sort_position_by_original(x_o, y_o, x_solved, y_solved):
    """Sorting new image positions such that the old order is best preserved.

    :param x_o: numpy array; original image positions
    :param y_o: numpy array; original image positions
    :param x_solved: numpy array; solved image positions with potentially more or fewer
        images
    :param y_solved: numpy array; solved image positions with potentially more or fewer
        images
    :return: sorted new image positions with the order best matching the original
        positions first, and then all other images in the same order as solved for
    """
    if len(x_o) > len(x_solved):
        # if new images are less , then return the original images (no sorting required)
        x_solved_new, y_solved_new = x_o, y_o
    else:
        x_solved_new, y_solved_new = [], []
        for i in range(len(x_o)):
            x, y = x_o[i], y_o[i]
            r2_i = (x - x_solved) ** 2 + (y - y_solved) ** 2
            # index of minimum radios
            index = np.argmin(r2_i)
            x_solved_new.append(x_solved[index])
            y_solved_new.append(y_solved[index])
            # delete this index
            x_solved = np.delete(x_solved, index)
            y_solved = np.delete(y_solved, index)
        # now we append the remaining additional images in the same order behind the original ones
        x_solved_new = np.append(np.array(x_solved_new), x_solved)
        y_solved_new = np.append(np.array(y_solved_new), y_solved)
    return x_solved_new, y_solved_new



from lenstronomy.GalKin import aperture_types
import pytest
import numpy as np


class TestApertureTypes(object):
    def setup_method(self):
        pass

    def test_shell_select(self):
        # aperture = Aperture()
        ra, dec = 1, 1
        r_in = 2
        r_out = 4
        bool_select = aperture_types.shell_select(
            ra, dec, r_in, r_out, center_ra=0, center_dec=0
        )
        assert bool_select is False

        bool_select = aperture_types.shell_select(
            3, 0, r_in, r_out, center_ra=0, center_dec=0
        )
        assert bool_select is True

    def test_slit_select(self):
        bool_select = aperture_types.slit_select(
            ra=0.9, dec=0, length=2, width=0.5, center_ra=0, center_dec=0, angle=0
        )
        assert bool_select is True

        bool_select = aperture_types.slit_select(
            ra=0.9,
            dec=0,
            length=2,
            width=0.5,
            center_ra=0,
            center_dec=0,
            angle=np.pi / 2,
        )
        assert bool_select is False

    def test_ifu_shell_select(self):
        ra, dec = 1, 1
        r_bin = np.linspace(0, 10, 11)
        bool_select, i = aperture_types.shell_ifu_select(
            ra, dec, r_bin, center_ra=0, center_dec=0
        )
        assert bool_select is True
        assert i == 1

    def test_frame(self):
        center_ra, center_dec = 0, 0
        width_outer = 1.2
        width_inner = 0.6
        ra, dec = 0, 0
        bool_select = aperture_types.frame_select(
            ra,
            dec,
            width_inner=width_inner,
            width_outer=width_outer,
            center_ra=center_ra,
            center_dec=center_dec,
            angle=0,
        )
        assert bool_select is False
        ra, dec = 0.5, 0
        bool_select = aperture_types.frame_select(
            ra,
            dec,
            width_inner=width_inner,
            width_outer=width_outer,
            center_ra=center_ra,
            center_dec=center_dec,
            angle=0,
        )
        assert bool_select is True
        ra, dec = 5, 5
        bool_select = aperture_types.frame_select(
            ra,
            dec,
            width_inner=width_inner,
            width_outer=width_outer,
            center_ra=center_ra,
            center_dec=center_dec,
            angle=0,
        )
        assert bool_select is False


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
import numpy as np

__all__ = ["FluxRatioLikelihood"]


class FluxRatioLikelihood(object):
    """Likelihood class for magnification of multiply lensed images."""

    def __init__(
        self,
        lens_model_class,
        flux_ratios,
        flux_ratio_errors,
        source_type="INF",
        window_size=0.1,
        grid_number=100,
        polar_grid=False,
        aspect_ratio=0.5,
    ):
        """

        :param lens_model_class: LensModel class instance
        :param flux_ratios: ratio of fluxes of the multiple images (relative to the first appearing in
         same order as the images)
        :param flux_ratio_errors: errors in the flux ratios (relative to the first appearing). Alternatively
         a log-normal covariance matrix. Note: in the case of a covariance matrix, the errors are assumed
         to be log-normal, i.e. the logarithms of the flux ratios, ln(F[i]/F[0]) are assumed to have a multivariate
         Gaussian distribution, with the given covariance matrix.
        :param source_type: string, type of source, 'INF' specifies a point source, while 'GAUSSIAN' specifies a
         finite-size source modeled as a Gaussian
        :param window_size: size of window to compute the finite flux
        :param grid_number: number of grid cells per axis in the window to numerically compute the flux
        """
        self._lens_model_class = lens_model_class
        self._flux_ratios = np.array(flux_ratios)
        self._flux_ratio_errors = np.array(flux_ratio_errors)
        self._lens_model_extensions = LensModelExtensions(lensModel=lens_model_class)
        self._source_type = source_type
        self._window_size = window_size
        self._gird_number = grid_number
        self._polar_grid = polar_grid
        self._aspect_ratio = aspect_ratio

    def logL(self, x_pos, y_pos, kwargs_lens, kwargs_special):
        """

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_special: dictionary of 'special' keyword parameters
        :return: log likelihood of the measured flux ratios given a model
        """
        if self._source_type == "INF":
            mag = np.abs(
                self._lens_model_class.magnification(x_pos, y_pos, kwargs_lens)
            )
        else:
            source_sigma = kwargs_special["source_size"]
            mag = self._lens_model_extensions.magnification_finite(
                x_pos,
                y_pos,
                kwargs_lens,
                source_sigma=source_sigma,
                window_size=self._window_size,
                grid_number=self._gird_number,
                polar_grid=self._polar_grid,
                aspect_ratio=self._aspect_ratio,
            )
        if len(mag) - 1 != len(self._flux_ratios):
            return -(10**15)
        mag_ratio = mag[1:] / mag[0]
        return self._logL(mag_ratio)

    def _logL(self, flux_ratios):
        """


        :param flux_ratios: flux ratios from the model
        :return: log likelihood fo the measured flux ratios given a model
        """
        if not np.isfinite(flux_ratios).any():
            return -(10**15)
        if self._flux_ratio_errors.ndim <= 1:
            dist = (
                (flux_ratios - self._flux_ratios) ** 2
                / self._flux_ratio_errors**2
                / 2
            )
            logL = -np.sum(dist)
        elif self._flux_ratio_errors.ndim == 2:
            # Assume covariance matrix is in ln units!
            D = np.log(flux_ratios) - np.log(self._flux_ratios)
            logL = (
                -1 / 2 * D @ np.linalg.inv(self._flux_ratio_errors) @ D
            )  # TODO: only calculate the inverse once
        else:
            raise ValueError(
                "flux_ratio_errors need dimension of 1 or 2. Current dimensions are %s"
                % self._flux_ratio_errors.ndim
            )
        if not np.isfinite(logL):
            return -(10**15)
        return logL

    @property
    def num_data(self):
        """

        :return: integer, number of data points associated with the flux ratios
        """
        return len(self._flux_ratios)



__all__ = ["MultiDataBase"]


class MultiDataBase(object):
    """Base class with definitions that are shared among all variations of modelling
    multiple data sets."""

    def __init__(self, image_model_list, compute_bool=None):
        """

        :param image_model_list: list of ImageModel instances (supporting linear inversions)
        :param compute_bool: list of booleans for each imaging band indicating whether to model it or not.
        """
        self._num_bands = len(image_model_list)
        if compute_bool is None:
            compute_bool = [True] * self._num_bands
        else:
            if not len(compute_bool) == self._num_bands:
                raise ValueError(
                    "compute_bool statement has not the same range as number of bands available!"
                )
        self._compute_bool = compute_bool
        self._imageModel_list = image_model_list
        self._num_response_list = []
        for imageModel in image_model_list:
            self._num_response_list.append(imageModel.num_data_evaluate)

    @property
    def num_bands(self):
        return self._num_bands

    @property
    def num_response_list(self):
        """List of number of data elements that are used in the minimization.

        :return: list of integers
        """
        return self._num_response_list

    def reset_point_source_cache(self, cache=True):
        """Deletes all the cache in the point source class and saves it from then on.

        :return:
        """
        for imageModel in self._imageModel_list:
            imageModel.reset_point_source_cache(cache=cache)

    @property
    def num_data_evaluate(self):
        num = 0
        for i in range(self._num_bands):
            if self._compute_bool[i] is True:
                num += self._imageModel_list[i].num_data_evaluate
        return num

    def num_param_linear(
        self, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
    ):
        """

        :return: number of linear coefficients to be solved for in the linear inversion
        """
        num = 0
        for i in range(self._num_bands):
            if self._compute_bool[i] is True:
                num += self._imageModel_list[i].num_param_linear(
                    kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
                )
        return num

    def reduced_residuals(self, model_list, error_map_list=None):
        """

        :param model_list: list of models
        :param error_map_list: list of error maps
        :return:
        """
        residual_list = []
        if error_map_list is None:
            error_map_list = [[] for _ in range(self._num_bands)]
        index = 0
        for i in range(self._num_bands):
            if self._compute_bool[i] is True:
                residual_list.append(
                    self._imageModel_list[i].reduced_residuals(
                        model_list[index], error_map=error_map_list[index]
                    )
                )
                index += 1
        return residual_list



import numpy as np
from lenstronomy.Conf import config_loader
from os import environ

"""
From pyautolens:
Depending on if we're using a super computer, we want two different numba decorators:
If on laptop:
@numba.jit(nopython=True, cache=True, parallel=False)
If on super computer:
@numba.jit(nopython=True, cache=False, parallel=True)
"""

numba_conf = config_loader.numba_conf()
nopython = numba_conf["nopython"]
cache = numba_conf["cache"]
parallel = numba_conf["parallel"]
numba_enabled = numba_conf["enable"] and not environ.get("NUMBA_DISABLE_JIT", False)
fastmath = numba_conf["fastmath"]
error_model = numba_conf["error_model"]

if numba_enabled:
    try:
        import numba
        from numba import extending
    except ImportError:
        numba_enabled = False
        numba = None
        extending = None

__all__ = ["jit", "overload", "nan_to_num", "nan_to_num_arr", "nan_to_num_single"]


def jit(
    nopython=nopython,
    cache=cache,
    parallel=parallel,
    fastmath=fastmath,
    error_model=error_model,
    inline="never",
):
    if numba_enabled:

        def wrapper(func):
            return numba.jit(
                func,
                nopython=nopython,
                cache=cache,
                parallel=parallel,
                fastmath=fastmath,
                error_model=error_model,
                inline=inline,
            )

    else:

        def wrapper(func):
            return func

    return wrapper


def overload(
    nopython=nopython,
    cache=cache,
    parallel=parallel,
    fastmath=fastmath,
    error_model=error_model,
):
    """Wrapper around numba.generated_jit.

    Allows you to redirect a function to another based on its type
    - see the Numba docs for more info
    """
    if numba_enabled:

        def wrapper(func):
            # TODO change to overload, but currently breaks tests with nopython
            return numba.generated_jit(
                func,
                nopython=nopython,
                cache=cache,
                parallel=parallel,
                fastmath=fastmath,
                error_model=error_model,
            )
            # return extending.overload(func, jit_options={'nopython': nopython, 'cache': cache,
            #                                             'parallel': parallel,
            #                                             'fastmath': fastmath, 'error_model': error_model})

    else:

        def wrapper(func):
            return func

    return wrapper


@overload()
def nan_to_num(x, posinf=1e10, neginf=-1e10, nan=0.0):
    """Implements a Numba equivalent to np.nan_to_num (with copy=False!) array or scalar
    in Numba.

    Behaviour is the same as np.nan_to_num with copy=False, although it only supports
    1-dimensional arrays and scalar inputs.
    """
    # The generated_jit part is necessary because of the need to support both arrays and scalars for all input
    # functions.
    if (
        (numba_enabled and isinstance(x, numba.types.Array))
        or isinstance(x, np.ndarray)
    ) and x.ndim > 0:
        return (
            nan_to_num_arr if numba_enabled else nan_to_num_arr(x, posinf, neginf, nan)
        )
    else:
        return (
            nan_to_num_single
            if numba_enabled
            else nan_to_num_single(x, posinf, neginf, nan)
        )


@jit()
def nan_to_num_arr(x, posinf=1e10, neginf=-1e10, nan=0.0):
    """Part of the Numba implementation of np.nan_to_num - see nan_to_num"""
    for i in range(len(x)):
        if np.isnan(x[i]):
            x[i] = nan
        if np.isinf(x[i]):
            if x[i] > 0:
                x[i] = posinf
            else:
                x[i] = neginf
    return x


@jit()
def nan_to_num_single(x, posinf=1e10, neginf=-1e10, nan=0.0):
    """Part of the Numba implementation of np.nan_to_num - see nan_to_num"""
    if np.isnan(x):
        return nan
    elif np.isinf(x):
        if x > 0:
            return posinf
        else:
            return neginf
    else:
        return x



__author__ = "lynevdv"

import numpy as np

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Multipole"]


class Multipole(LensProfileBase):
    """
    This class contains a multipole contribution (for 1 component with m>=2)
    This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf
    Equation B12
    m : int, multipole order, m>=2
    a_m : float, multipole strength
    phi_m : float, multipole orientation in radian
    """

    param_names = ["m", "a_m", "phi_m", "center_x", "center_y"]
    lower_limit_default = {
        "m": 2,
        "a_m": 0,
        "phi_m": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "m": 100,
        "a_m": 100,
        "phi_m": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Lensing potential of multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: lensing potential
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        f_ = r * a_m / (1 - m**2) * np.cos(m * (phi - phi_m))
        return f_

    def derivatives(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Deflection of a multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf
        Equation B12

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: deflection angles alpha_x, alpha_y
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        f_x = np.cos(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) + np.sin(
            phi
        ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))
        f_y = np.sin(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) - np.cos(
            phi
        ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))
        return f_x, f_y

    def hessian(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Hessian of a multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: f_xx, f_xy, f_yx, f_yy
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        r = np.maximum(r, 0.000001)
        f_xx = 1.0 / r * np.sin(phi) ** 2 * a_m * np.cos(m * (phi - phi_m))
        f_yy = 1.0 / r * np.cos(phi) ** 2 * a_m * np.cos(m * (phi - phi_m))
        f_xy = -1.0 / r * a_m * np.cos(phi) * np.sin(phi) * np.cos(m * (phi - phi_m))
        return f_xx, f_xy, f_xy, f_yy



from lenstronomy.Util.magnification_finite_util import setup_mag_finite
import numpy as np
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions


def plot_quasar_images(
    lens_model,
    x_image,
    y_image,
    source_x,
    source_y,
    kwargs_lens,
    source_fwhm_parsec,
    z_source,
    cosmo=None,
    grid_resolution=None,
    grid_radius_arcsec=None,
    source_light_model="SINGLE_GAUSSIAN",
    dx=None,
    dy=None,
    size_scale=None,
    amp_scale=None,
):
    """This function plots the surface brightness in the image plane of a background
    source modeled as either a single Gaussian or two Gaussian light profiles. The flux
    is computed inside a circular aperture with radius grid_radius_arcsec. If
    grid_radius_arcsec is not specified a default value will be assumed.

    :param lens_model: an instance of LensModel
    :param x_image: a list or array of x coordinates [units arcsec]
    :param y_image: a list or array of y coordinates [units arcsec]
    :param kwargs_lens: keyword arguments for the lens model
    :param source_fwhm_parsec: the size of the background source [units parsec]
    :param z_source: the source redshift
    :param cosmo: (optional) an instance of astropy.cosmology; if not specified, a
        default cosmology will be used
    :param grid_resolution: the grid resolution in units arcsec/pixel; if not specified,
        an appropriate value will be estimated from the source size
    :param grid_radius_arcsec: (optional) the size of the ray tracing region in arcsec;
        if not specified, an appropriate value will be estimated from the source size
    :param source_light_model: the model for background source light; currently
        implemented are 'SINGLE_GAUSSIAN' and 'DOUBLE_GAUSSIAN'.
    :param dx: used with source model 'DOUBLE_GAUSSIAN', the offset of the second source
        light profile from the first [arcsec]
    :param dy: used with source model 'DOUBLE_GAUSSIAN', the offset of the second source
        light profile from the first [arcsec]
    :param size_scale: used with source model 'DOUBLE_GAUSSIAN', the size of the second
        source light profile relative to the first
    :param amp_scale: used with source model 'DOUBLE_GAUSSIAN', the peak brightness of
        the second source light profile relative to the first
    :return: Four images of the background source in the image plane
    """

    lens_model_extension = LensModelExtensions(lens_model)

    magnifications = []
    images = []

    (
        grid_x_0,
        grid_y_0,
        source_model,
        kwargs_source,
        grid_resolution,
        grid_radius_arcsec,
    ) = setup_mag_finite(
        cosmo,
        lens_model,
        grid_radius_arcsec,
        grid_resolution,
        source_fwhm_parsec,
        source_light_model,
        z_source,
        source_x,
        source_y,
        dx,
        dy,
        amp_scale,
        size_scale,
    )
    shape0 = grid_x_0.shape
    grid_x_0, grid_y_0 = grid_x_0.ravel(), grid_y_0.ravel()

    for xi, yi in zip(x_image, y_image):
        flux_array = np.zeros_like(grid_x_0)
        r_min = 0
        r_max = grid_radius_arcsec
        grid_r = np.hypot(grid_x_0, grid_y_0)
        flux_array = lens_model_extension._magnification_adaptive_iteration(
            flux_array,
            xi,
            yi,
            grid_x_0,
            grid_y_0,
            grid_r,
            r_min,
            r_max,
            lens_model,
            kwargs_lens,
            source_model,
            kwargs_source,
        )
        m = np.sum(flux_array) * grid_resolution**2
        magnifications.append(m)
        images.append(flux_array.reshape(shape0))

    magnifications = np.array(magnifications)
    flux_ratios = magnifications / max(magnifications)
    import matplotlib.pyplot as plt

    fig = plt.figure(1)
    fig.set_size_inches(16, 6)
    N = len(images)
    for i, (image, mag, fr) in enumerate(zip(images, magnifications, flux_ratios)):
        ax = plt.subplot(1, N, i + 1)
        ax.imshow(
            image,
            origin="lower",
            extent=[
                -grid_radius_arcsec,
                grid_radius_arcsec,
                -grid_radius_arcsec,
                grid_radius_arcsec,
            ],
        )
        ax.annotate(
            "magnification: " + str(np.round(mag, 3)),
            xy=(0.05, 0.9),
            xycoords="axes fraction",
            color="w",
            fontsize=12,
        )
        ax.annotate(
            "flux ratio: " + str(np.round(fr, 3)),
            xy=(0.05, 0.8),
            xycoords="axes fraction",
            color="w",
            fontsize=12,
        )
    plt.show()



import numpy as np
from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil
import numpy.testing as npt


def test_bn():
    n_array = np.linspace(start=0.2, stop=8, num=30)
    for n in n_array:
        bn_approx = 1.9992 * n - 0.3271
        bn = SersicUtil.b_n(n)
        npt.assert_almost_equal(bn, bn_approx, decimal=3)



__author__ = "sibirrer"

import pytest
import unittest
import numpy as np
import numpy.testing as npt
from lenstronomy.PointSource.point_source_param import PointSourceParam


class TestParam(object):
    def setup_method(self):
        kwargs_fixed = [{}, {}, {}]
        num_point_sources_list = [4, 1, 1]
        fixed_magnification_list = [True, False, False]
        point_source_model_list = ["LENSED_POSITION", "SOURCE_POSITION", "UNLENSED"]
        self.param = PointSourceParam(
            model_list=point_source_model_list,
            kwargs_fixed=kwargs_fixed,
            num_point_source_list=num_point_sources_list,
            fixed_magnification_list=fixed_magnification_list,
        )
        self.kwargs = [
            {
                "ra_image": np.array([0, 0, 0, 0]),
                "dec_image": np.array([0, 0, 0, 0]),
                "source_amp": 1,
            },
            {"ra_source": 1, "dec_source": 1, "point_amp": np.array([1.0])},
            {"ra_image": [1], "dec_image": [1], "point_amp": np.array([1.0])},
        ]

        self.param_linear = PointSourceParam(
            model_list=point_source_model_list,
            kwargs_fixed=[{}, {}, {}],
            num_point_source_list=num_point_sources_list,
            linear_solver=False,
            fixed_magnification_list=fixed_magnification_list,
        )

    def test_get_setParams(self):
        args = self.param.set_params(self.kwargs)
        kwargs_new, _ = self.param.get_params(args, i=0)
        args_new = self.param.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

        args = self.param_linear.set_params(self.kwargs)
        kwargs_new, _ = self.param_linear.get_params(args, i=0)
        args_new = self.param_linear.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

    def test_num_params(self):
        num, list = self.param.num_param()
        assert num == 12

        num, list = self.param_linear.num_param()
        assert num == 15

    def test_num_param_linear(self):
        num = self.param.num_param_linear()
        assert num == 3

        num = self.param_linear.num_param_linear()
        assert num == 0

    def test_init(self):
        ps_param = PointSourceParam(
            model_list=["UNLENSED"], kwargs_fixed=[{}], num_point_source_list=None
        )
        assert ps_param._num_point_sources_list[0] == 1


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            PointSourceParam(
                model_list=["BAD"],
                kwargs_fixed=[{}],
                kwargs_lower=None,
                kwargs_upper=[{"bla": 1}],
            )
        with self.assertRaises(ValueError):
            PointSourceParam(
                model_list=["BAD"],
                kwargs_fixed=[{}],
                kwargs_lower=[{"bla": 1}],
                kwargs_upper=None,
            )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.single_plane import SinglePlane
from lenstronomy.LensModel.Profiles.sis import SIS
import unittest

try:
    import fastell4py

    bool_test = True
except:
    bool_test = False


class TestLensModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.lensModel = SinglePlane(["GAUSSIAN"])
        self.kwargs = [
            {
                "amp": 1.0,
                "sigma_x": 2.0,
                "sigma_y": 2.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]

    def test_potential(self):
        output = self.lensModel.potential(x=1.0, y=1.0, kwargs=self.kwargs)
        assert output == 0.77880078307140488 / (8 * np.pi)

    def test_alpha(self):
        output1, output2 = self.lensModel.alpha(x=1.0, y=1.0, kwargs=self.kwargs)
        assert output1 == -0.19470019576785122 / (8 * np.pi)
        assert output2 == -0.19470019576785122 / (8 * np.pi)

    def test_ray_shooting(self):
        delta_x, delta_y = self.lensModel.ray_shooting(x=1.0, y=1.0, kwargs=self.kwargs)
        assert delta_x == 1 + 0.19470019576785122 / (8 * np.pi)
        assert delta_y == 1 + 0.19470019576785122 / (8 * np.pi)

    def test_mass_2d(self):
        lensModel = SinglePlane(["GAUSSIAN_KAPPA"])
        kwargs = [{"amp": 1.0, "sigma": 2.0, "center_x": 0.0, "center_y": 0.0}]
        output = lensModel.mass_2d(r=1, kwargs=kwargs)
        npt.assert_almost_equal(output, 0.11750309741540453, decimal=9)

    def test_density(self):
        theta_E = 1
        r = 1
        lensModel = SinglePlane(lens_model_list=["SIS"])
        density = lensModel.density(r=r, kwargs=[{"theta_E": theta_E}])
        sis = SIS()
        density_model = sis.density_lens(r=r, theta_E=theta_E)
        npt.assert_almost_equal(density, density_model, decimal=8)

    def test_bool_list(self):
        lensModel = SinglePlane(["SPEP", "SHEAR"])
        kwargs = [
            {
                "theta_E": 1,
                "gamma": 2,
                "e1": 0.1,
                "e2": -0.1,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.01, "gamma2": -0.02},
        ]
        alphax_1, alphay_1 = lensModel.alpha(1, 1, kwargs, k=0)
        alphax_1_list, alphay_1_list = lensModel.alpha(1, 1, kwargs, k=[0])
        npt.assert_almost_equal(alphax_1, alphax_1_list, decimal=5)
        npt.assert_almost_equal(alphay_1, alphay_1_list, decimal=5)

        alphax_1_1, alphay_1_1 = lensModel.alpha(1, 1, kwargs, k=0)
        alphax_1_2, alphay_1_2 = lensModel.alpha(1, 1, kwargs, k=1)
        alphax_full, alphay_full = lensModel.alpha(1, 1, kwargs, k=None)
        npt.assert_almost_equal(alphax_1_1 + alphax_1_2, alphax_full, decimal=5)
        npt.assert_almost_equal(alphay_1_1 + alphay_1_2, alphay_full, decimal=5)

    def test_init(self):
        lens_model_list = [
            "TNFW",
            "TRIPLE_CHAMELEON",
            "SHEAR_GAMMA_PSI",
            "CURVED_ARC_CONST",
            "NFW_MC",
            "NFW_MC_ELLIPSE",
            "ARC_PERT",
            "MULTIPOLE",
            "CURVED_ARC_SPP",
        ]
        lensModel = SinglePlane(lens_model_list=lens_model_list)
        assert lensModel.func_list[0].param_names[0] == "Rs"


class TestRaise(unittest.TestCase):
    def test_raise(self):
        """Check whether raises occurs if fastell4py is not installed.

        :return:
        """
        if bool_test is False:
            with self.assertRaises(ImportError):
                SinglePlane(lens_model_list=["PEMD"])
            with self.assertRaises(ImportError):
                SinglePlane(lens_model_list=["SPEMD"])
        else:
            SinglePlane(lens_model_list=["PEMD", "SPEMD"])


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "sibirrer"

# description of the polar shapelets in potential space

import numpy as np
import scipy.special
import math

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["PolarShapelets"]


class PolarShapelets(LensProfileBase):
    """This class contains the function and the derivatives of the Singular Isothermal
    Sphere."""

    param_names = ["coeffs", "beta", "center_x", "center_y"]
    lower_limit_default = {"coeffs": [0], "beta": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {
        "coeffs": [100],
        "beta": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        n = 10
        self.poly = [[[] for i in range(n)] for j in range(n)]
        for i in range(0, n):
            for j in range(0, n):
                self.poly[i][j] = scipy.special.genlaguerre(i, j)
        super(PolarShapelets, self).__init__()

    def function(self, x, y, coeffs, beta, center_x=0, center_y=0):
        shapelets = self._createShapelet(coeffs)
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        f_ = self._shapeletOutput(r, phi, beta, shapelets)
        return f_

    def derivatives(self, x, y, coeffs, beta, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        shapelets = self._createShapelet(coeffs)
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        alpha1_shapelets, alpha2_shapelets = self._alphaShapelets(shapelets, beta)
        f_x = self._shapeletOutput(r, phi, beta, alpha1_shapelets)
        f_y = self._shapeletOutput(r, phi, beta, alpha2_shapelets)
        return f_x, f_y

    def hessian(self, x, y, coeffs, beta, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        shapelets = self._createShapelet(coeffs)
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        kappa_shapelets = self._kappaShapelets(shapelets, beta)
        gamma1_shapelets, gamma2_shapelets = self._gammaShapelets(shapelets, beta)
        kappa_value = self._shapeletOutput(r, phi, beta, kappa_shapelets)
        gamma1_value = self._shapeletOutput(r, phi, beta, gamma1_shapelets)
        gamma2_value = self._shapeletOutput(r, phi, beta, gamma2_shapelets)
        f_xx = kappa_value + gamma1_value
        f_xy = gamma2_value
        f_yy = kappa_value - gamma1_value
        return f_xx, f_xy, f_xy, f_yy

    def _createShapelet(self, coeff):
        """Returns a shapelet array out of the coefficients *a, up to order l.

        :param num_l: order of shapelets
        :type num_l: int.
        :param coeff: shapelet coefficients
        :type coeff: floats
        :returns: complex array
        :raises: AttributeError, KeyError
        """
        n_coeffs = len(coeff)
        num_l = self._get_num_l(n_coeffs)
        shapelets = np.zeros((num_l + 1, num_l + 1), "complex")
        nl = 0
        k = 0
        i = 0
        while i < len(coeff):
            if i % 2 == 0:
                shapelets[nl][k] += coeff[i] / 2.0
                shapelets[k][nl] += coeff[i] / 2.0
                if k == nl:
                    nl += 1
                    k = 0
                    i += 1
                    continue
                else:
                    k += 1
                    i += 1
                    continue
            else:
                shapelets[nl][k] += 1j * coeff[i] / 2.0
                shapelets[k][nl] -= 1j * coeff[i] / 2.0
                i += 1
        return shapelets

    def _shapeletOutput(self, r, phi, beta, shapelets):
        """Returns the the numerical values of a set of shapelets at polar coordinates.

        :param shapelets: set of shapelets [l=,r=,a_lr=]
        :type shapelets: array of size (n,3)
        :param coordPolar: set of coordinates in polar units
        :type coordPolar: array of size (n,2)
        :returns: array of same size with coords [r,phi]
        :raises: AttributeError, KeyError
        """
        if (
            type(r) == float
            or type(r) == int
            or type(r) == type(np.float64(1))
            or len(r) <= 1
        ):
            values = 0.0
        else:
            values = np.zeros(len(r), "complex")
        for nl in range(0, len(shapelets)):  # sum over different shapelets
            for nr in range(0, len(shapelets)):
                value = shapelets[nl][nr] * self._chi_lr(r, phi, nl, nr, beta)
                values += value
        return values.real

    def _chi_lr(self, r, phi, nl, nr, beta):
        """Computes the generalized polar basis function in the convention of
        Massey&Refregier eqn 8.

        :param nl: left basis
        :type nl: int
        :param nr: right basis
        :type nr: int
        :param beta: beta --the characteristic scale typically choosen to be close to
            the size of the object.
        :type beta: float.
        :param coord: coordinates [r,phi]
        :type coord: array(n,2)
        :returns: values at positions of coordinates.
        :raises: AttributeError, KeyError
        """
        m = int((nr - nl).real)
        n = int((nr + nl).real)
        p = int((n - abs(m)) / 2)
        p2 = int((n + abs(m)) / 2)
        q = int(abs(m))
        if p % 2 == 0:  # if p is even
            prefac = 1
        else:
            prefac = -1
        prefactor = (
            prefac
            / beta ** (abs(m) + 1)
            * np.sqrt(math.factorial(p) / (np.pi * math.factorial(p2)))
        )
        poly = self.poly[p][q]
        return (
            prefactor
            * r**q
            * poly((r / beta) ** 2)
            * np.exp(-((r / beta) ** 2) / 2)
            * np.exp(-1j * m * phi)
        )

    def _kappaShapelets(self, shapelets, beta):
        """Calculates the convergence kappa given lensing potential shapelet
        coefficients (laplacian/2) :param shapelets: set of shapelets [l=,r=,a_lr=]
        :type shapelets: array of size (n,3) :returns:  set of kappa shapelets.

        :raises: AttributeError, KeyError
        """
        output = np.zeros((len(shapelets) + 1, len(shapelets) + 1), "complex")
        for nl in range(0, len(shapelets)):
            for nr in range(0, len(shapelets)):
                a_lr = shapelets[nl][nr]
                if nl > 0:
                    output[nl - 1][nr + 1] += a_lr * np.sqrt(nl * (nr + 1)) / 2
                    if nr > 0:
                        output[nl - 1][nr - 1] += a_lr * np.sqrt(nl * nr) / 2
                output[nl + 1][nr + 1] += a_lr * np.sqrt((nl + 1) * (nr + 1)) / 2
                if nr > 0:
                    output[nl + 1][nr - 1] += a_lr * np.sqrt((nl + 1) * nr) / 2
        return output / beta**2

    def _alphaShapelets(self, shapelets, beta):
        """Calculates the deflection angles given lensing potential shapelet
        coefficients (laplacian/2) :param shapelets: set of shapelets [l=,r=,a_lr=]
        :type shapelets: array of size (n,3) :returns:  set of alpha shapelets.

        :raises: AttributeError, KeyError
        """
        output_x = np.zeros((len(shapelets) + 1, len(shapelets) + 1), "complex")
        output_y = np.zeros((len(shapelets) + 1, len(shapelets) + 1), "complex")
        for nl in range(0, len(shapelets)):
            for nr in range(0, len(shapelets)):
                a_lr = shapelets[nl][nr]
                output_x[nl][nr + 1] -= a_lr * np.sqrt(nr + 1) / 2
                output_y[nl][nr + 1] -= a_lr * np.sqrt(nr + 1) / 2 * 1j
                output_x[nl + 1][nr] -= a_lr * np.sqrt(nl + 1) / 2
                output_y[nl + 1][nr] += a_lr * np.sqrt(nl + 1) / 2 * 1j
                if nl > 0:
                    output_x[nl - 1][nr] += a_lr * np.sqrt(nl) / 2
                    output_y[nl - 1][nr] -= a_lr * np.sqrt(nl) / 2 * 1j
                if nr > 0:
                    output_x[nl][nr - 1] += a_lr * np.sqrt(nr) / 2
                    output_y[nl][nr - 1] += a_lr * np.sqrt(nr) / 2 * 1j
        return output_x / beta, output_y / beta  # attention complex numbers!!!!

    def _gammaShapelets(self, shapelets, beta):
        """Calculates the shear gamma given lensing potential shapelet coefficients
        :param shapelets: set of shapelets [l=,r=,a_lr=] :type shapelets: array of size
        (n,3) :returns:  set of alpha shapelets.

        :raises: AttributeError, KeyError
        """
        output_x = np.zeros((len(shapelets) + 2, len(shapelets) + 2), "complex")
        output_y = np.zeros((len(shapelets) + 2, len(shapelets) + 2), "complex")
        for nl in range(0, len(shapelets)):
            for nr in range(0, len(shapelets)):
                a_lr = shapelets[nl][nr]
                output_x[nl + 2][nr] += a_lr * np.sqrt((nl + 1) * (nl + 2)) / 2
                output_x[nl][nr + 2] += a_lr * np.sqrt((nr + 1) * (nr + 2)) / 2
                output_x[nl][nr] += a_lr * (1 - (nr + 1) - (nl + 1))
                if nl > 1:
                    output_x[nl - 2][nr] += a_lr * np.sqrt((nl) * (nl - 1)) / 2
                if nr > 1:
                    output_x[nl][nr - 2] += a_lr * np.sqrt((nr) * (nr - 1)) / 2

                output_y[nl + 2][nr] += a_lr * np.sqrt((nl + 1) * (nl + 2)) * 1j / 4
                output_y[nl][nr + 2] -= a_lr * np.sqrt((nr + 1) * (nr + 2)) * 1j / 4
                if nl > 0:
                    output_y[nl - 1][nr + 1] += a_lr * np.sqrt((nl) * (nr + 1)) * 1j / 2
                if nr > 0:
                    output_y[nl + 1][nr - 1] -= a_lr * np.sqrt((nr) * (nl + 1)) * 1j / 2
                if nl > 1:
                    output_y[nl - 2][nr] -= a_lr * np.sqrt((nl) * (nl - 1)) * 1j / 4
                if nr > 1:
                    output_y[nl][nr - 2] += a_lr * np.sqrt((nr) * (nr - 1)) * 1j / 4
        return (
            output_x / beta**2,
            output_y / beta**2,
        )  # attention complex numbers!!!!

    @staticmethod
    def _get_num_l(n_coeffs):
        """

        :param n_coeffs: number of coeffs
        :return: number of n_l of order of the shapelets
        """
        num_l = int(round((math.sqrt(8 * n_coeffs + 9) - 3) / 2 + 0.499))
        return num_l



from lenstronomy.Workflow.multi_band_manager import MultiBandUpdateManager
from lenstronomy.Workflow.update_manager import UpdateManager
import pytest


class TestUpdateManager(object):
    def setup_method(self):
        kwargs_model = {
            "lens_model_list": ["SHEAR", "SHEAR"],
            "source_light_model_list": ["UNIFORM"],
            "lens_light_model_list": ["UNIFORM"],
            "optical_depth_model_list": [],
        }
        kwargs_constraints = {}
        kwargs_likelihood = {}
        kwargs_params = {}
        lens_init = [{"e1": 0, "e2": 0}, {"e1": 0, "e2": 0}]
        lens_sigma = [{"e1": 0.1, "e2": 0.1}, {"e1": 0.1, "e2": 0.1}]
        lens_fixed = [{"ra_0": 0, "dec_0": 0}, {"ra_0": 0, "dec_0": 0}]
        lens_lower = [{"e1": -1, "e2": -1}, {"e1": -1, "e2": -1}]
        lens_upper = [{"e1": 1, "e2": 1}, {"e1": 1, "e2": 1}]
        kwargs_params["lens_model"] = [
            lens_init,
            lens_sigma,
            lens_fixed,
            lens_lower,
            lens_upper,
        ]
        kwargs_params["source_model"] = [[{}], [{}], [{}], [{}], [{}]]
        kwargs_params["lens_light_model"] = [[{}], [{}], [{}], [{}], [{}]]
        kwargs_params["special"] = [
            {"special1": 1},
            {"special1": 1},
            {"special1": 0.1},
            {"special1": 0},
            {"special1": 1},
        ]
        kwargs_params["extinction_model"] = [[], [], [], [], []]
        self.manager = UpdateManager(
            kwargs_model, kwargs_constraints, kwargs_likelihood, kwargs_params
        )

    def test_none_mamager(self):
        manager = MultiBandUpdateManager(
            kwargs_model={},
            kwargs_constraints={},
            kwargs_likelihood={},
            kwargs_params={},
            num_bands=0,
        )
        results = manager.best_fit()
        assert len(results["kwargs_lens"]) == 0

    def test_init_kwargs(self):
        kwargs_init = self.manager.init_kwargs
        assert kwargs_init["kwargs_lens"][0]["e1"] == 0

    def test_sigma_kwargs(self):
        kwargs_sigma = self.manager.sigma_kwargs
        assert kwargs_sigma["kwargs_lens"][0]["e1"] == 0.1

    def test_update_kwargs_model(self):
        kwargs_model = {
            "lens_model_list": ["SHEAR", "SHEAR"],
            "source_light_model_list": ["UNIFORM"],
            "lens_light_model_list": ["UNIFORM"],
            "optical_depth_model_list": [],
            "multi_plane": True,
            "z_source": 2.0,
            "lens_redshift_list": [0.5, 0.6],
            "source_redshift_list": [2.0],
        }
        kwargs_constraints = {
            "lens_redshift_sampling_indexes": [-1, 1],
            "source_redshift_sampling_indexes": [0],
        }
        kwargs_likelihood = {}
        kwargs_params = {}
        lens_init = [{"e1": 0, "e2": 0}, {"e1": 0, "e2": 0}]
        lens_sigma = [{"e1": 0.1, "e2": 0.1}, {"e1": 0.1, "e2": 0.1}]
        lens_fixed = [{"ra_0": 0, "dec_0": 0}, {"ra_0": 0, "dec_0": 0}]
        lens_lower = [{"e1": -1, "e2": -1}, {"e1": -1, "e2": -1}]
        lens_upper = [{"e1": 1, "e2": 1}, {"e1": 1, "e2": 1}]
        kwargs_params["lens_model"] = [
            lens_init,
            lens_sigma,
            lens_fixed,
            lens_lower,
            lens_upper,
        ]
        kwargs_params["source_model"] = [[{}], [{}], [{}], [{}], [{}]]
        kwargs_params["lens_light_model"] = [[{}], [{}], [{}], [{}], [{}]]
        kwargs_params["special"] = [
            {"special1": 1},
            {"special1": 1},
            {"special1": 0.1},
            {"special1": 0},
            {"special1": 1},
        ]
        kwargs_params["extinction_model"] = [[], [], [], [], []]

        manager = UpdateManager(
            kwargs_model, kwargs_constraints, kwargs_likelihood, kwargs_params
        )

        kwargs_special = {"z_sampling": [1.9, 0.7]}
        manager.update_kwargs_model(kwargs_special)

        assert manager.kwargs_model["lens_redshift_list"] == [0.5, 0.7]
        assert manager.kwargs_model["source_redshift_list"] == [1.9]

    def test_update_parameter_state(self):
        self.manager.update_param_state(
            kwargs_lens=[{"e1": -2, "e2": 0}, {"e1": 2, "e2": 0}]
        )
        kwargs_temp = self.manager.parameter_state
        assert kwargs_temp["kwargs_lens"][0]["e1"] == -2
        self.manager.set_init_state()
        kwargs_temp = self.manager.parameter_state
        assert kwargs_temp["kwargs_lens"][0]["e1"] == 0

    def test_update_param_value(self):
        self.manager.update_param_value(lens=[[1, ["e1"], [0.029]]])
        kwargs_temp = self.manager.parameter_state
        assert kwargs_temp["kwargs_lens"][1]["e1"] == 0.029

    def test_param_class(self):
        param_class = self.manager.param_class
        num_param, param_names = param_class.num_param()
        assert num_param == 4

    def test_best_fit(self):
        kwargs_result = self.manager.best_fit(bijective=True)
        assert kwargs_result["kwargs_lens"][0]["e1"] == 0

    def test_update_options(self):
        self.manager.update_options(
            kwargs_model=None,
            kwargs_constraints={"test": "test"},
            kwargs_likelihood=None,
        )
        assert self.manager.kwargs_constraints["test"] == "test"

        self.manager.update_options(
            kwargs_model={"test": "test"},
            kwargs_constraints=None,
            kwargs_likelihood=None,
        )
        assert self.manager.kwargs_model["test"] == "test"

    def test_update_limits(self):
        self.manager.update_limits(
            change_source_lower_limit=[[0, ["test"], [-1]]],
            change_source_upper_limit=[[0, ["test"], [1]]],
        )
        self.manager.update_limits(
            change_lens_lower_limit=[[0, ["e1"], [-0.9]]],
            change_lens_upper_limit=[[0, ["e1"], [0.9]]],
        )
        upper_lens, upper_source, _, _, _, _ = self.manager._upper_kwargs
        assert upper_source[0]["test"] == 1
        assert upper_lens[0]["e1"] == 0.9

    def test_update_sigmas(self):
        self.manager.update_sigmas(
            change_sigma_source=[[0, ["test"], [1]]],
            change_sigma_lens=[[0, ["test"], [2]]],
        )
        self.manager.update_sigmas(
            change_sigma_lens_light=[[0, ["e1"], [-0.9]]],
            change_sigma_lens=[[0, ["e1"], [0.9]]],
        )
        upper_lens, upper_source, _, _, _, _ = self.manager._upper_kwargs
        assert self.manager._lens_sigma[0]["test"] == 2
        assert self.manager._lens_sigma[0]["e1"] == 0.9

    def test_update_fixed(self):
        lens_add_fixed = [[0, ["e1"], [-1]]]
        self.manager.update_fixed(lens_add_fixed=lens_add_fixed)
        assert self.manager._lens_fixed[0]["e1"] == -1

        lens_add_fixed = [[0, ["e2"]]]
        self.manager.update_fixed(lens_add_fixed=lens_add_fixed)
        assert self.manager._lens_fixed[0]["e2"] == 0

        lens_remove_fixed = [[0, ["e1"]]]
        self.manager.update_fixed(lens_remove_fixed=lens_remove_fixed)
        assert "e1" not in self.manager._lens_fixed[0]

        assert "special1" in self.manager._special_fixed
        self.manager.update_fixed(special_remove_fixed=["special1"])
        assert "special1" not in self.manager._special_fixed

        self.manager.update_fixed(special_add_fixed=["special1"])
        assert self.manager._special_fixed["special1"] == 1

        self.manager.update_fixed(special_add_fixed=["special1"])
        assert self.manager._special_fixed["special1"] == 1

    def test_update_logsampling(self):
        self.manager.update_options(
            kwargs_model={},
            kwargs_constraints={"log_sampling_lens": [[0, ["e1"]]]},
            kwargs_likelihood={},
        )
        assert self.manager.param_class.lensParams.kwargs_logsampling[0] == ["e1"]

    def test_fix_image_parameters(self):
        self.manager.fix_image_parameters(image_index=0)
        assert 1 == 1


if __name__ == "__main__":
    pytest.main()



from lenstronomy.ImSim.MultiBand.multi_data_base import MultiDataBase
from lenstronomy.ImSim.MultiBand.single_band_multi_model import SingleBandMultiModel

__all__ = ["MultiLinear"]


class MultiLinear(MultiDataBase):
    """Class to simulate/reconstruct images in multi-band option. This class calls
    functions of image_model.py with different bands with joint non-linear parameters
    and decoupled linear parameters.

    the class supports keyword arguments 'index_lens_model_list', 'index_source_light_model_list',
    'index_lens_light_model_list', 'index_point_source_model_list', 'index_optical_depth_model_list' in kwargs_model
    These arguments should be lists of length the number of imaging bands available and each entry in the list
    is a list of integers specifying the model components being evaluated for the specific band.

    E.g. there are two bands and you want to different light profiles being modeled.
    - you define two different light profiles lens_light_model_list = ['SERSIC', 'SERSIC']
    - set index_lens_light_model_list = [[0], [1]]
    - (optional) for now all the parameters between the two light profiles are independent in the model. You have
    the possibility to join a subset of model parameters (e.g. joint centroid). See the Param() class for documentation.
    """

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        likelihood_mask_list=None,
        compute_bool=None,
        kwargs_pixelbased=None,
        linear_solver=True,
    ):
        """

        :param multi_band_list: list of imaging band configurations [[kwargs_data, kwargs_psf, kwargs_numerics],[...], ...]
        :param kwargs_model: model option keyword arguments
        :param likelihood_mask_list: list of likelihood masks (booleans with size of the individual images)
        :param compute_bool: (optional), bool list to indicate which band to be included in the modeling
        :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
         that they get overwritten by the linear solver solution.
        """
        self.type = "multi-linear"
        imageModel_list = []
        if linear_solver is False and len(multi_band_list) > 1:
            raise ValueError(
                'Multi-linear mode with more than one band does not support "linear_solver" = False.'
            )
        for band_index in range(len(multi_band_list)):
            imageModel = SingleBandMultiModel(
                multi_band_list,
                kwargs_model,
                likelihood_mask_list=likelihood_mask_list,
                band_index=band_index,
                kwargs_pixelbased=kwargs_pixelbased,
                linear_solver=linear_solver,
            )
            imageModel_list.append(imageModel)
        super(MultiLinear, self).__init__(imageModel_list, compute_bool=compute_bool)

    def image_linear_solve(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        inv_bool=False,
    ):
        """Computes the image (lens and source surface brightness with a given lens
        model). The linear parameters are computed with a weighted linear least square
        optimization (i.e. flux normalization of the brightness profiles)

        :param kwargs_lens: list of keyword arguments corresponding to the superposition
            of different lens profiles
        :param kwargs_source: list of keyword arguments corresponding to the
            superposition of different source light profiles
        :param kwargs_lens_light: list of keyword arguments corresponding to different
            lens light surface brightness profiles
        :param kwargs_ps: keyword arguments corresponding to "other" parameters, such as
            external shear and point source image positions
        :param inv_bool: if True, invert the full linear solver Matrix Ax = y for the
            purpose of the covariance matrix.
        :return: 1d array of surface brightness pixels of the optimal solution of the
            linear parameters to match the data
        """
        wls_list, error_map_list, cov_param_list, param_list = [], [], [], []
        for i in range(self._num_bands):
            if self._compute_bool[i] is True:
                wls_model, error_map, cov_param, param = self._imageModel_list[
                    i
                ].image_linear_solve(
                    kwargs_lens,
                    kwargs_source,
                    kwargs_lens_light,
                    kwargs_ps,
                    kwargs_extinction,
                    kwargs_special,
                    inv_bool=inv_bool,
                )
            else:
                wls_model, error_map, cov_param, param = None, None, None, None
            wls_list.append(wls_model)
            error_map_list.append(error_map)
            cov_param_list.append(cov_param)
            param_list.append(param)
        return wls_list, error_map_list, cov_param_list, param_list

    def likelihood_data_given_model(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_extinction=None,
        kwargs_special=None,
        source_marg=False,
        linear_prior=None,
        check_positive_flux=False,
    ):
        """Computes the likelihood of the data given a model This is specified with the
        non-linear parameters and a linear inversion and prior marginalisation.

        :param kwargs_lens:
        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :param check_positive_flux: bool, if True, checks whether the linear inversion
            resulted in non-negative flux components and applies a punishment in the
            likelihood if so.
        :return: log likelihood (natural logarithm) (sum of the log likelihoods of the
            individual images)
        """
        # generate image
        logL = 0
        if linear_prior is None:
            linear_prior = [None for i in range(self._num_bands)]
        for i in range(self._num_bands):
            if self._compute_bool[i] is True:
                logL += self._imageModel_list[i].likelihood_data_given_model(
                    kwargs_lens,
                    kwargs_source,
                    kwargs_lens_light,
                    kwargs_ps,
                    kwargs_extinction,
                    kwargs_special,
                    source_marg=source_marg,
                    linear_prior=linear_prior[i],
                    check_positive_flux=check_positive_flux,
                )
        return logL



import copy
from lenstronomy.Sampling.parameters import Param

__all__ = ["UpdateManager"]


class UpdateManager(object):
    """This class manages the parameter constraints as they may evolve through the steps
    of the modeling.

    This includes: keeping certain parameters fixed during one modelling step
    """

    def __init__(
        self, kwargs_model, kwargs_constraints, kwargs_likelihood, kwargs_params
    ):
        """

        :param kwargs_model: keyword arguments to describe all model components used in
         class_creator.create_class_instances()
        :param kwargs_constraints: keyword arguments of the Param() class to handle parameter constraints during the
         sampling (except upper and lower limits and sampling input mean and width)
        :param kwargs_likelihood: keyword arguments of the Likelihood() class to handle parameters and settings of the
         likelihood
        :param kwargs_params: setting of the sampling bounds and initial guess mean and spread.
         The argument is organized as:
         'lens_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'source_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'lens_light_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'point_source_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'extinction_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'special': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]

        """
        self.kwargs_model = kwargs_model
        self.kwargs_constraints = kwargs_constraints
        self.kwargs_likelihood = kwargs_likelihood

        if kwargs_model.get("lens_model_list", None) is not None:
            (
                self._lens_init,
                self._lens_sigma,
                self._lens_fixed,
                self._lens_lower,
                self._lens_upper,
            ) = kwargs_params["lens_model"]
        else:
            (
                self._lens_init,
                self._lens_sigma,
                self._lens_fixed,
                self._lens_lower,
                self._lens_upper,
            ) = ([], [], [], [], [])
        if kwargs_model.get("source_light_model_list", None) is not None:
            (
                self._source_init,
                self._source_sigma,
                self._source_fixed,
                self._source_lower,
                self._source_upper,
            ) = kwargs_params["source_model"]
        else:
            (
                self._source_init,
                self._source_sigma,
                self._source_fixed,
                self._source_lower,
                self._source_upper,
            ) = ([], [], [], [], [])
        if kwargs_model.get("lens_light_model_list", None) is not None:
            (
                self._lens_light_init,
                self._lens_light_sigma,
                self._lens_light_fixed,
                self._lens_light_lower,
                self._lens_light_upper,
            ) = kwargs_params["lens_light_model"]
        else:
            (
                self._lens_light_init,
                self._lens_light_sigma,
                self._lens_light_fixed,
                self._lens_light_lower,
                self._lens_light_upper,
            ) = ([], [], [], [], [])
        if kwargs_model.get("point_source_model_list", None) is not None:
            (
                self._ps_init,
                self._ps_sigma,
                self._ps_fixed,
                self._ps_lower,
                self._ps_upper,
            ) = kwargs_params["point_source_model"]
        else:
            (
                self._ps_init,
                self._ps_sigma,
                self._ps_fixed,
                self._ps_lower,
                self._ps_upper,
            ) = ([], [], [], [], [])
        if kwargs_model.get("optical_depth_model_list", None) is not None:
            (
                self._extinction_init,
                self._extinction_sigma,
                self._extinction_fixed,
                self._extinction_lower,
                self._extinction_upper,
            ) = kwargs_params["extinction_model"]
        else:
            (
                self._extinction_init,
                self._extinction_sigma,
                self._extinction_fixed,
                self._extinction_lower,
                self._extinction_upper,
            ) = ([], [], [], [], [])
        if "special" in kwargs_params:
            (
                self._special_init,
                self._special_sigma,
                self._special_fixed,
                self._special_lower,
                self._special_upper,
            ) = kwargs_params["special"]
        else:
            (
                self._special_init,
                self._special_sigma,
                self._special_fixed,
                self._special_lower,
                self._special_upper,
            ) = ({}, {}, {}, {}, {})

        self._kwargs_temp = self.init_kwargs

    # TODO: check compatibility with number of point sources provided as well as other parameter labeling
    # TODO: give raise statement if sigma value is =0

    @property
    def init_kwargs(self):
        """

        :return: keyword arguments for all model components of the initial mean model proposition in the sampling
        """
        return {
            "kwargs_lens": self._lens_init,
            "kwargs_source": self._source_init,
            "kwargs_lens_light": self._lens_light_init,
            "kwargs_ps": self._ps_init,
            "kwargs_special": self._special_init,
            "kwargs_extinction": self._extinction_init,
        }

    @property
    def sigma_kwargs(self):
        """

        :return: keyword arguments for all model components of the initial 1-sigma width proposition in the sampling
        """
        return {
            "kwargs_lens": self._lens_sigma,
            "kwargs_source": self._source_sigma,
            "kwargs_lens_light": self._lens_light_sigma,
            "kwargs_ps": self._ps_sigma,
            "kwargs_special": self._special_sigma,
            "kwargs_extinction": self._extinction_sigma,
        }

    @property
    def _lower_kwargs(self):
        return (
            self._lens_lower,
            self._source_lower,
            self._lens_light_lower,
            self._ps_lower,
            self._special_lower,
            self._extinction_lower,
        )

    @property
    def _upper_kwargs(self):
        return (
            self._lens_upper,
            self._source_upper,
            self._lens_light_upper,
            self._ps_upper,
            self._special_upper,
            self._extinction_upper,
        )

    @property
    def fixed_kwargs(self):
        return (
            self._lens_fixed,
            self._source_fixed,
            self._lens_light_fixed,
            self._ps_fixed,
            self._special_fixed,
            self._extinction_fixed,
        )

    def set_init_state(self):
        """Set the current state of the parameters to the initial one.

        :return:
        """
        self._kwargs_temp = self.init_kwargs

    @property
    def parameter_state(self):
        """

        :return: parameter state saved in this class
        """
        return self._kwargs_temp

    def best_fit(self, bijective=False):
        """Best fit (max likelihood) position for all the model parameters.

        :param bijective: boolean, if True, returns the parameters in the argument of
            the sampling that might deviate from the convention of the ImSim module. For
            example, if parameterized in the image position, the parameters remain in
            the image plane rather than being mapped to the source plane.
        :return: kwargs_result with all the keyword arguments of the best fit for the
            model components
        """
        (
            lens_temp,
            source_temp,
            lens_light_temp,
            ps_temp,
            special_temp,
            extinction_temp,
        ) = (
            self._kwargs_temp["kwargs_lens"],
            self._kwargs_temp["kwargs_source"],
            self._kwargs_temp["kwargs_lens_light"],
            self._kwargs_temp["kwargs_ps"],
            self._kwargs_temp["kwargs_special"],
            self._kwargs_temp["kwargs_extinction"],
        )
        if bijective is False:
            lens_temp = self.param_class.update_lens_scaling(
                special_temp, lens_temp, inverse=False
            )
            source_temp = self.param_class.image2source_plane(source_temp, lens_temp)
        return {
            "kwargs_lens": lens_temp,
            "kwargs_source": source_temp,
            "kwargs_lens_light": lens_light_temp,
            "kwargs_ps": ps_temp,
            "kwargs_special": special_temp,
            "kwargs_extinction": extinction_temp,
        }

    def update_param_state(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
    ):
        """Updates the temporary state of the parameters being saved. ATTENTION: Any
        previous knowledge gets lost if you call this function.

        :param kwargs_lens:
        :param kwargs_source:
        :param kwargs_lens_light:
        :param kwargs_ps:
        :param kwargs_special:
        :param kwargs_extinction:
        :return:
        """
        self._kwargs_temp = {
            "kwargs_lens": kwargs_lens,
            "kwargs_source": kwargs_source,
            "kwargs_lens_light": kwargs_lens_light,
            "kwargs_ps": kwargs_ps,
            "kwargs_special": kwargs_special,
            "kwargs_extinction": kwargs_extinction,
        }
        self.update_kwargs_model(kwargs_special)

    def update_param_value(self, lens=None, source=None, lens_light=None, ps=None):
        """Set a model parameter to a specific value.

        :param lens: [[i_model, ['param1', 'param2',...], [...]]
        :param source: [[i_model, ['param1', 'param2',...], [...]]
        :param lens_light: [[i_model, ['param1', 'param2',...], [...]]
        :param ps: [[i_model, ['param1', 'param2',...], [...]]
        :return: 0, the value of the param is overwritten
        """
        if lens is None:
            lens = []
        if source is None:
            source = []
        if lens_light is None:
            lens_light = []
        if ps is None:
            ps = []
        for items, kwargs_key in zip(
            [lens, source, lens_light, ps],
            ["kwargs_lens", "kwargs_source", "kwargs_lens_light", "kwargs_ps"],
        ):
            for item in items:
                index = item[0]
                keys = item[1]
                values = item[2]

                for key, value in zip(keys, values):
                    self._kwargs_temp[kwargs_key][index][key] = value

    @property
    def param_class(self):
        """Creating instance of lenstronomy Param() class. It uses the keyword arguments
        in self.kwargs_constraints as __init__() arguments, as well as
        self.kwargs_model, and the set of kwargs_fixed___, kwargs_lower___,
        kwargs_upper___ arguments for lens, lens_light, source, point source, extinction
        and special parameters.

        :return: instance of the Param class with the recent options and bounds
        """
        (
            kwargs_fixed_lens,
            kwargs_fixed_source,
            kwargs_fixed_lens_light,
            kwargs_fixed_ps,
            kwargs_fixed_special,
            kwargs_fixed_extinction,
        ) = self.fixed_kwargs
        (
            kwargs_lower_lens,
            kwargs_lower_source,
            kwargs_lower_lens_light,
            kwargs_lower_ps,
            kwargs_lower_special,
            kwargs_lower_extinction,
        ) = self._lower_kwargs
        (
            kwargs_upper_lens,
            kwargs_upper_source,
            kwargs_upper_lens_light,
            kwargs_upper_ps,
            kwargs_upper_special,
            kwargs_upper_extinction,
        ) = self._upper_kwargs
        kwargs_model = self.kwargs_model
        kwargs_constraints = self.kwargs_constraints
        lens_temp = self._kwargs_temp["kwargs_lens"]
        param_class = Param(
            kwargs_model,
            kwargs_fixed_lens,
            kwargs_fixed_source,
            kwargs_fixed_lens_light,
            kwargs_fixed_ps,
            kwargs_fixed_special,
            kwargs_fixed_extinction,
            kwargs_lower_lens,
            kwargs_lower_source,
            kwargs_lower_lens_light,
            kwargs_lower_ps,
            kwargs_lower_special,
            kwargs_lower_extinction,
            kwargs_upper_lens,
            kwargs_upper_source,
            kwargs_upper_lens_light,
            kwargs_upper_ps,
            kwargs_upper_special,
            kwargs_upper_extinction,
            kwargs_lens_init=lens_temp,
            **kwargs_constraints
        )
        return param_class

    def update_kwargs_model(self, kwargs_special):
        """Update the kwargs_model with the new kwargs_special."""
        kwargs_model, update_bool = self.param_class.update_kwargs_model(kwargs_special)
        if update_bool:
            self.kwargs_model = kwargs_model

        return kwargs_model

    def update_options(
        self, kwargs_model=None, kwargs_constraints=None, kwargs_likelihood=None
    ):
        """
        updates the options by overwriting the kwargs with the new ones being added/changed
        WARNING: some updates may not be valid depending on the model options. Use carefully!

        :param kwargs_model: keyword arguments to describe all model components used in
         class_creator.create_class_instances() that are updated from previous arguments
        :param kwargs_constraints:
        :param kwargs_likelihood:
        :return: kwargs_model, kwargs_constraints, kwargs_likelihood
        """
        if kwargs_model is None:
            kwargs_model = {}
        if kwargs_constraints is None:
            kwargs_constraints = {}
        if kwargs_likelihood is None:
            kwargs_likelihood = []
        kwargs_model_updated = self.kwargs_model.update(kwargs_model)
        kwargs_constraints_updated = self.kwargs_constraints.update(kwargs_constraints)
        kwargs_likelihood_updated = self.kwargs_likelihood.update(kwargs_likelihood)
        return (
            kwargs_model_updated,
            kwargs_constraints_updated,
            kwargs_likelihood_updated,
        )

    def update_limits(
        self,
        change_source_lower_limit=None,
        change_source_upper_limit=None,
        change_lens_lower_limit=None,
        change_lens_upper_limit=None,
    ):
        """Updates the limits (lower and upper) of the update manager instance.

        :param change_source_lower_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_lens_lower_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_source_upper_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_lens_upper_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :return: updates internal state of lower and upper limits accessible from outside
        """
        if change_source_lower_limit is not None:
            self._source_lower = self._update_kwargs_list(
                change_source_lower_limit, self._source_lower
            )
        if change_source_upper_limit is not None:
            self._source_upper = self._update_kwargs_list(
                change_source_upper_limit, self._source_upper
            )
        if change_lens_lower_limit is not None:
            self._lens_lower = self._update_kwargs_list(
                change_lens_lower_limit, self._lens_lower
            )
        if change_lens_upper_limit is not None:
            self._lens_upper = self._update_kwargs_list(
                change_lens_upper_limit, self._lens_upper
            )

    def update_sigmas(
        self,
        change_sigma_lens=None,
        change_sigma_source=None,
        change_sigma_lens_light=None,
    ):
        """Updates individual estimated uncertainty levels for the initialization of
        search and sampling algorithms.

        :param change_sigma_lens: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_sigma_source: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_sigma_lens_light: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :return: updated internal state of the spread to initialize samplers
        """
        if change_sigma_lens is not None:
            self._lens_sigma = self._update_kwargs_list(
                change_sigma_lens, self._lens_sigma
            )
        if change_sigma_source is not None:
            self._source_sigma = self._update_kwargs_list(
                change_sigma_source, self._source_sigma
            )
        if change_sigma_lens_light is not None:
            self._lens_light_sigma = self._update_kwargs_list(
                change_sigma_lens_light, self._lens_light_sigma
            )

    @staticmethod
    def _update_kwargs_list(change_list, kwargs_list_previous):
        """

        :param change_list: input format of def update_limits [[i_model, ['param_name1', 'param_name2', ...],
         [value1, value2, ...]]]
        :param kwargs_list_previous: keyword argument list
        :return: update limits
        """
        kwargs_limit_updated = copy.deepcopy(kwargs_list_previous)
        for i in range(len(change_list)):
            i_model = change_list[i][0]
            change_names = change_list[i][1]
            values = change_list[i][2]
            for j, param_name in enumerate(change_names):
                kwargs_limit_updated[i_model][param_name] = values[j]
        return kwargs_limit_updated

    def update_fixed(
        self,
        lens_add_fixed=None,
        source_add_fixed=None,
        lens_light_add_fixed=None,
        ps_add_fixed=None,
        special_add_fixed=None,
        lens_remove_fixed=None,
        source_remove_fixed=None,
        lens_light_remove_fixed=None,
        ps_remove_fixed=None,
        special_remove_fixed=None,
    ):
        """Adds or removes the values of the keyword arguments that are stated in the
        _add_fixed to the existing fixed arguments. convention for input arguments are:
        [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...
        (optional)], [], ...]

        :param lens_add_fixed: added fixed parameter in lens model
        :param source_add_fixed: added fixed parameter in source model
        :param lens_light_add_fixed: added fixed parameter in lens light model
        :param ps_add_fixed: added fixed parameter in point source model
        :param special_add_fixed: added fixed parameter in special model
        :param lens_remove_fixed: remove fixed parameter in lens model
        :param source_remove_fixed: remove fixed parameter in source model
        :param lens_light_remove_fixed: remove fixed parameter in lens light model
        :param ps_remove_fixed: remove fixed parameter in point source model
        :param special_remove_fixed: remove fixed parameter in special model
        :return: updated kwargs fixed
        """
        lens_fixed = self._add_fixed(
            self._kwargs_temp["kwargs_lens"], self._lens_fixed, lens_add_fixed
        )
        lens_fixed = self._remove_fixed(lens_fixed, lens_remove_fixed)
        source_fixed = self._add_fixed(
            self._kwargs_temp["kwargs_source"], self._source_fixed, source_add_fixed
        )
        source_fixed = self._remove_fixed(source_fixed, source_remove_fixed)
        lens_light_fixed = self._add_fixed(
            self._kwargs_temp["kwargs_lens_light"],
            self._lens_light_fixed,
            lens_light_add_fixed,
        )
        lens_light_fixed = self._remove_fixed(lens_light_fixed, lens_light_remove_fixed)
        ps_fixed = self._add_fixed(
            self._kwargs_temp["kwargs_ps"], self._ps_fixed, ps_add_fixed
        )
        ps_fixed = self._remove_fixed(ps_fixed, ps_remove_fixed)
        special_fixed = copy.deepcopy(self._special_fixed)
        special_temp = self._kwargs_temp["kwargs_special"]
        if special_add_fixed is None:
            special_add_fixed = []
        for param_name in special_add_fixed:
            if param_name not in special_fixed:
                special_fixed[param_name] = special_temp[param_name]
        if special_remove_fixed is None:
            special_remove_fixed = []
        for param_name in special_remove_fixed:
            if param_name in special_fixed:
                del special_fixed[param_name]
        (
            self._lens_fixed,
            self._source_fixed,
            self._lens_light_fixed,
            self._ps_fixed,
            self._special_fixed,
        ) = (lens_fixed, source_fixed, lens_light_fixed, ps_fixed, special_fixed)

    @staticmethod
    def _add_fixed(kwargs_model, kwargs_fixed, add_fixed):
        """

        :param kwargs_model: model parameters
        :param kwargs_fixed: parameters that are held fixed (even before)
        :param add_fixed: additional fixed parameters
         [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ... (optional)], [], ...]
        :return: updated kwargs_fixed
        """
        if add_fixed is None:
            add_fixed = []
        # fixed_kwargs = copy.deepcopy(kwargs_fixed)
        for i in range(len(add_fixed)):
            i_model = add_fixed[i][0]
            fix_names = add_fixed[i][1]
            if len(add_fixed[i]) > 2:
                values = add_fixed[i][2]
            else:
                values = [None] * len(fix_names)
            for j, param_name in enumerate(fix_names):
                if values[j] is None:
                    kwargs_fixed[i_model][param_name] = kwargs_model[i_model][
                        param_name
                    ]  # add fixed list
                else:
                    kwargs_fixed[i_model][param_name] = values[j]
        return kwargs_fixed

    @staticmethod
    def _remove_fixed(kwargs_fixed, remove_fixed):
        """

        :param kwargs_fixed: fixed parameters (before)
        :param remove_fixed: list of parameters to be removed from the fixed list and initialized by the valuye of
         kwargs_model [[i_model, ['param_name1', 'param_name2', ...]], [], ...]
        :return: updated kwargs fixed parameters
        """
        if remove_fixed is None:
            remove_fixed = []
        for i in range(len(remove_fixed)):
            i_model = remove_fixed[i][0]
            fix_names = remove_fixed[i][1]
            for param_name in fix_names:
                if (
                    param_name in kwargs_fixed[i_model]
                ):  # if the parameter already is in the fixed list, do not change it
                    del kwargs_fixed[i_model][param_name]
        return kwargs_fixed

    def fix_image_parameters(self, image_index=0):
        """Fixes all parameters that are only assigned to a specific image. This allows
        to sample only parameters that constraint by the fitting of a sub-set of the
        images.

        :param image_index: index
        :return: None
        """
        pass



from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np

__all__ = ["SIE"]


class SIE(LensProfileBase):
    """Class for singular isothermal ellipsoid (SIS with ellipticity)

    .. math::
        \\kappa(x, y) = \\frac{1}{2} \\left(\\frac{\\theta_{E}}{\\sqrt{q x^2 + y^2/q}} \\right)

    with :math:`\\theta_{E}` is the (circularized) Einstein radius,
    :math:`q` is the minor/major axis ratio,
    and :math:`x` and :math:`y` are defined in a coordinate sys- tem aligned with the major and minor axis of the lens.

    In terms of eccentricities, this profile is defined as

    .. math::
        \\kappa(r) = \\frac{1}{2} \\left(\\frac{\\theta'_{E}}{r \\sqrt{1 − e*\\cos(2*\\phi)}} \\right)

    with :math:`\\epsilon` is the ellipticity defined as

    .. math::
        \\epsilon = \\frac{1-q^2}{1+q^2}

    And an Einstein radius :math:`\\theta'_{\\rm E}` related to the definition used is

    .. math::
        \\left(\\frac{\\theta'_{\\rm E}}{\\theta_{\\rm E}}\\right)^{2} = \\frac{2q}{1+q^2}.
    """

    param_names = ["theta_E", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, NIE=True):
        """

        :param NIE: bool, if True, is using the NIE analytic model. Otherwise it uses PEMD with gamma=2 from fastell4py
        """
        self._nie = NIE
        if NIE:
            from lenstronomy.LensModel.Profiles.nie import NIE

            self.profile = NIE()
        else:
            from lenstronomy.LensModel.Profiles.epl import EPL

            self.profile = EPL()
        self._s_scale = 0.0000000001
        self._gamma = 2
        super(SIE, self).__init__()

    def function(self, x, y, theta_E, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate (angular coordinates)
        :param y: y-coordinate (angular coordinates)
        :param theta_E: Einstein radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: centroid
        :param center_y: centroid
        :return:
        """
        if self._nie:
            return self.profile.function(
                x, y, theta_E, e1, e2, self._s_scale, center_x, center_y
            )
        else:
            return self.profile.function(
                x, y, theta_E, self._gamma, e1, e2, center_x, center_y
            )

    def derivatives(self, x, y, theta_E, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate (angular coordinates)
        :param y: y-coordinate (angular coordinates)
        :param theta_E: Einstein radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: centroid
        :param center_y: centroid
        :return:
        """
        if self._nie:
            return self.profile.derivatives(
                x, y, theta_E, e1, e2, self._s_scale, center_x, center_y
            )
        else:
            return self.profile.derivatives(
                x, y, theta_E, self._gamma, e1, e2, center_x, center_y
            )

    def hessian(self, x, y, theta_E, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinate (angular coordinates)
        :param y: y-coordinate (angular coordinates)
        :param theta_E: Einstein radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: centroid
        :param center_y: centroid
        :return:
        """
        if self._nie:
            return self.profile.hessian(
                x, y, theta_E, e1, e2, self._s_scale, center_x, center_y
            )
        else:
            return self.profile.hessian(
                x, y, theta_E, self._gamma, e1, e2, center_x, center_y
            )

    @staticmethod
    def theta2rho(theta_E):
        """Converts projected density parameter (in units of deflection) into 3d density
        parameter.

        :param theta_E:
        :return:
        """
        fac1 = np.pi * 2
        rho0 = theta_E / fac1
        return rho0

    @staticmethod
    def mass_3d(r, rho0, e1=0, e2=0):
        """Mass enclosed a 3d sphere or radius r.

        :param r: radius in angular units
        :param rho0: density at angle=1
        :return: mass in angular units
        """
        mass_3d = 4 * np.pi * rho0 * r
        return mass_3d

    def mass_3d_lens(self, r, theta_E, e1=0, e2=0):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units.

        :param r: radius in angular units
        :param theta_E: Einstein radius
        :return: mass in angular units
        """
        rho0 = self.theta2rho(theta_E)
        return self.mass_3d(r, rho0)

    def mass_2d(self, r, rho0, e1=0, e2=0):
        """Mass enclosed projected 2d sphere of radius r.

        :param r:
        :param rho0:
        :param e1:
        :param e2:
        :return:
        """
        alpha = 2 * rho0 * np.pi**2
        mass_2d = alpha * r
        return mass_2d

    def mass_2d_lens(self, r, theta_E, e1=0, e2=0):
        """

        :param r:
        :param theta_E:
        :param e1:
        :param e2:
        :return:
        """
        rho0 = self.theta2rho(theta_E)
        return self.mass_2d(r, rho0)

    def grav_pot(self, x, y, rho0, e1=0, e2=0, center_x=0, center_y=0):
        """Gravitational potential (modulo 4 pi G and rho0 in appropriate units)

        :param x:
        :param y:
        :param rho0:
        :param e1:
        :param e2:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        mass_3d = self.mass_3d(r, rho0)
        pot = mass_3d / r
        return pot

    def density_lens(self, r, theta_E, e1=0, e2=0):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: radius in angles
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: density
        """
        rho0 = self.theta2rho(theta_E)
        return self.density(r, rho0)

    @staticmethod
    def density(r, rho0, e1=0, e2=0):
        """Computes the density.

        :param r: radius in angles
        :param rho0: density at angle=1
        :return: density at r
        """
        rho = rho0 / r**2
        return rho

    @staticmethod
    def density_2d(x, y, rho0, e1=0, e2=0, center_x=0, center_y=0):
        """Projected density.

        :param x:
        :param y:
        :param rho0:
        :param e1:
        :param e2:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        sigma = np.pi * rho0 / r
        return sigma



import numpy as np
import numpy.testing as npt
from lenstronomy.Sampling.Likelihoods.kinematic_2D_likelihood import KinLikelihood
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Data.psf import PSF
from lenstronomy.Data.kinematic_bin_2D import KinBin
from lenstronomy.Data.pixel_grid import PixelGrid
import lenstronomy.Util.kernel_util as kernel_util
from lenstronomy.Sampling.Likelihoods import kinematic_NN_call


class TestKinLikelihood(object):
    def setup(self):
        self.kinematic_NN = kinematic_NN_call.KinematicNN().SKiNN_installed
        # initialize KinLikelihood class and calculate example vrms for testing
        major_axis_pa = 30 * np.pi / 180
        mass_q = 0.7
        light_q = 0.6
        center_x = 0.0
        center_y = 0.0
        self.kwargs_lens = [
            {
                "theta_E": 1.5,
                "gamma": 2,
                "q": mass_q,
                "phi": major_axis_pa,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        self.kwargs_lens_light = [
            {
                "amp": 10,
                "R_sersic": 1.0,
                "q": light_q,
                "phi": major_axis_pa,
                "n_sersic": 3.0,
                "center_x": center_x,
                "center_y": center_y,
            }
        ]
        self.kwargs_special = {
            "D_dt": 4000,
            "b_ani": 0.2,
            "incli": np.pi / 2,
            "D_d": 1000,
        }

        lens_model_list = ["EPL_Q_PHI"]
        self.lensModel = LensModel(lens_model_list=lens_model_list, z_lens=0.5)
        lens_light_model_list = ["SERSIC_ELLIPSE_Q_PHI"]

        self.lensLightModel = LightModel(lens_light_model_list)
        npix = 3
        delta_pix_image = 0.2
        transform_pix2angle = np.array([[1, 0], [0, 1]]) * delta_pix_image
        ra_at_xy_0 = -(npix - 1) / 2.0 * delta_pix_image
        dec_at_xy_0 = -(npix - 1) / 2.0 * delta_pix_image

        image_pixel_grid = PixelGrid(
            npix, npix, transform_pix2angle, ra_at_xy_0, dec_at_xy_0
        )
        self.image_data = self.lensLightModel.surface_brightness(
            image_pixel_grid._x_grid, image_pixel_grid._y_grid, self.kwargs_lens_light
        )

        binmap = np.array([[0, 0, 1], [0, 0, 1], [2, 2, 3]])
        binned_dummy_data = np.ones(4)
        delta_pix_kin = delta_pix_image
        npix_kin = npix

        # create PSF for kinematics
        sharp_kin_psf = kernel_util.kernel_gaussian(
            num_pix=npix, delta_pix=delta_pix_image, fwhm=0.001
        )

        # KinLikelihood inputs: light image kwargs, kinematic image kwargs, PSF, lens model class, light model class
        self.kwargs_data = {
            "image_data": self.image_data,
            "noise_map": 100 * np.ones((npix, npix)),
            "transform_pix2angle": transform_pix2angle,
            "ra_at_xy_0": ra_at_xy_0,
            "dec_at_xy_0": dec_at_xy_0,
        }
        self.kwargs_kin = {
            "bin_data": binned_dummy_data,
            "bin_cov": np.diag((binned_dummy_data * 0.05) ** 2),  # 5% error
            "bin_mask": binmap,
            "ra_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
            "dec_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
            "transform_pix2angle": np.array([[1, 0], [0, 1]]) * delta_pix_kin,
        }

        kwargs_pixelkin = {"psf_type": "PIXEL", "kernel_point_source": sharp_kin_psf}
        self.kinPSF = PSF(**kwargs_pixelkin)
        _KinBin = KinBin(psf_class=self.kinPSF, **self.kwargs_kin)
        self._KinLikelihood = KinLikelihood(
            _KinBin,
            self.lensModel,
            self.lensLightModel,
            self.kwargs_data,
            idx_lens=0,
            idx_lens_light=0,
        )
        if self.kinematic_NN:
            # if NN enabled, calculate realistic values to use as truth for logL tests
            self.truth_vrms = self._KinLikelihood.calc_vrms(
                self.kwargs_lens,
                self.kwargs_lens_light,
                self.kwargs_special,
                verbose=False,
            )
            self.kwargs_kin = {
                "bin_data": self.truth_vrms,
                "bin_cov": np.diag((self.truth_vrms * 0.05) ** 2),  # 5% error
                "bin_mask": binmap,
                "ra_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
                "dec_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
                "transform_pix2angle": np.array([[1, 0], [0, 1]]) * delta_pix_kin,
            }
            _KinBin = KinBin(psf_class=self.kinPSF, **self.kwargs_kin)
            self._KinLikelihood = KinLikelihood(
                _KinBin,
                self.lensModel,
                self.lensLightModel,
                self.kwargs_data,
                idx_lens=0,
                idx_lens_light=0,
            )

    def test_logL(self):
        if self.kinematic_NN:
            kwargs_lens_close = [{}]
            kwargs_lens_far = [{}]
            for key, val in self.kwargs_lens[0].items():
                kwargs_lens_close[0][key] = val + 0.01
                kwargs_lens_far[0][key] = val + 0.2
            kwargs_lens_light_close = [{}]
            kwargs_lens_light_far = [{}]
            for key, val in self.kwargs_lens_light[0].items():
                kwargs_lens_light_close[0][key] = val + 0.01
                kwargs_lens_light_far[0][key] = val + 0.2

            close_logL = self._KinLikelihood.logL(
                kwargs_lens_close,
                kwargs_lens_light_close,
                self.kwargs_special,
                verbose=False,
            )
            far_logL = self._KinLikelihood.logL(
                kwargs_lens_far,
                kwargs_lens_light_far,
                self.kwargs_special,
                verbose=False,
            )
            assert (
                close_logL > -10
            )  # instance close to truth should be likely (usually around -0.5)
            assert (
                close_logL > far_logL
            )  # param instance close to truth is more likely than far away

            # test  that doubling variance decreases logL by factor of 2
            kwargs_kin = self.kwargs_kin.copy()
            kwargs_kin["bin_cov"] = self.kwargs_kin["bin_cov"] * 2
            _KinBin = KinBin(psf_class=self.kinPSF, **kwargs_kin)
            var_kin_likelihood = KinLikelihood(
                _KinBin,
                self.lensModel,
                self.lensLightModel,
                self.kwargs_data,
                idx_lens=0,
                idx_lens_light=0,
            )
            var_logL = var_kin_likelihood.logL(
                kwargs_lens_close,
                kwargs_lens_light_close,
                self.kwargs_special,
                verbose=False,
            )
            npt.assert_almost_equal(var_logL, close_logL / 2, decimal=2)

            # test that if velocity off by 1 sigma, logL off by 1 sigma
            kwargs_kin = self.kwargs_kin.copy()
            kwargs_kin["bin_data"][-1] = (
                kwargs_kin["bin_data"][-1] * 1.05
            )  # increase one bin by 1 sigma

            _KinBin = KinBin(psf_class=self.kinPSF, **kwargs_kin)
            sig_kin_likelihood = KinLikelihood(
                _KinBin,
                self.lensModel,
                self.lensLightModel,
                self.kwargs_data,
                idx_lens=0,
                idx_lens_light=0,
            )
            sig_logL = sig_kin_likelihood.logL(
                kwargs_lens_close,
                kwargs_lens_light_close,
                self.kwargs_special,
                verbose=False,
            )
            npt.assert_almost_equal(sig_logL, close_logL - 1 / 2, decimal=1)
            kwargs_out_of_bounds = kwargs_lens_close.copy()
            kwargs_out_of_bounds[0]["theta_E"] = 10
            out_of_bounds_logL = self._KinLikelihood.logL(
                kwargs_lens_close,
                kwargs_lens_light_close,
                self.kwargs_special,
                verbose=False,
            )
            assert out_of_bounds_logL == -(10**8)

    def test_convert_to_nn_params(self):
        kwargs_lens_test = [
            {
                "theta_E": 2.0,
                "gamma": 2.0,
                "q": 1.0,
                "phi": 0,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.06, "gamma2": -0.03},
        ]
        kwargs_lens_light_test = [
            {
                "amp": 10,
                "R_sersic": 1.0,
                "q": 1,
                "phi": 0,
                "n_sersic": 3.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]
        params, same_orientation = self._KinLikelihood.convert_to_nn_params(
            kwargs_lens_test, kwargs_lens_light_test, self.kwargs_special
        )
        npt.assert_array_equal(
            params,
            np.array(
                [
                    1.0,
                    1.0,
                    2.0,
                    3.0,
                    1.0,
                    8.0e-2,
                    0.5,
                    self.kwargs_special["b_ani"],
                    self.kwargs_special["incli"] * 180 / np.pi,
                ]
            ),
        )
        # check that wrong orientations are caught
        assert same_orientation == True
        kwargs_lens_test = [
            {
                "theta_E": 2.0,
                "gamma": 2.0,
                "q": 0.5,
                "phi": 1.0,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.06, "gamma2": -0.03},
        ]
        params, same_orientation = self._KinLikelihood.convert_to_nn_params(
            kwargs_lens_test, kwargs_lens_light_test, self.kwargs_special
        )
        assert same_orientation == False

    def test_rescale_distance(self):
        # scale=fiducial
        kwargs_special = {"D_dt": 2886.544, "b_ani": 0.1, "incli": 0.0, "D_d": 1215.739}
        rescaled_map = self._KinLikelihood.rescale_distance(
            self.image_data, kwargs_special
        )
        npt.assert_allclose(rescaled_map, self.image_data, atol=10**-4)
        # scale D_dt
        kwargs_special = {
            "D_dt": 2886.544 * 2,
            "b_ani": 0.1,
            "incli": 0.0,
            "D_d": 1215.739,
        }
        rescaled_map = self._KinLikelihood.rescale_distance(
            self.image_data, kwargs_special
        )
        npt.assert_allclose(np.sqrt(2) * self.image_data, rescaled_map, atol=10**-4)
        # scale D_d
        kwargs_special = {
            "D_dt": 2886.544,
            "b_ani": 0.1,
            "incli": 0.0,
            "D_d": 1215.739 * 2,
        }
        rescaled_map = self._KinLikelihood.rescale_distance(
            self.image_data, kwargs_special
        )
        npt.assert_allclose(
            1 / np.sqrt(2) * self.image_data, rescaled_map, atol=10**-4
        )

    def test_convert_kwargs_to_kinnalign_input(self):
        self._KinLikelihood.update_image_input(
            self.kwargs_lens
        )  # adds PA and centers to kwargs
        assert "ellipse_PA" in self._KinLikelihood.image_input

    def test_auto_binning(self):
        # light weighting testing, change light image
        dummy_vrms_map = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        sharp_image = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]]) + 0.0001

        # auto_binning inherits the psf, bin map, and data shape
        sharp_kin_psf = kernel_util.kernel_gaussian(num_pix=5, delta_pix=0.2, fwhm=0.01)
        kwargs_pixelkin = {"psf_type": "PIXEL", "kernel_point_source": sharp_kin_psf}
        kin_psf = PSF(**kwargs_pixelkin)
        _KinBin = KinBin(psf_class=kin_psf, **self.kwargs_kin)
        _KinLikelihood = KinLikelihood(
            _KinBin,
            self.lensModel,
            self.lensLightModel,
            self.kwargs_data,
            idx_lens=0,
            idx_lens_light=0,
        )
        vrms = _KinLikelihood.auto_binning(dummy_vrms_map, sharp_image)
        npt.assert_allclose(vrms, [5, 4.5, 7.5, 9], rtol=1e-3)
        vrms = _KinLikelihood.auto_binning(dummy_vrms_map, np.ones((3, 3)))
        npt.assert_allclose(vrms, [3, 4.5, 7.5, 9], rtol=1e-3)

        wide_kin_psf = kernel_util.kernel_gaussian(num_pix=5, delta_pix=0.2, fwhm=100)
        kwargs_pixelkin = {"psf_type": "PIXEL", "kernel_point_source": wide_kin_psf}
        kin_psf = PSF(**kwargs_pixelkin)
        _KinBin = KinBin(psf_class=kin_psf, **self.kwargs_kin)
        _KinLikelihood = KinLikelihood(
            _KinBin,
            self.lensModel,
            self.lensLightModel,
            self.kwargs_data,
            idx_lens=0,
            idx_lens_light=0,
        )

        vrms = _KinLikelihood.auto_binning(dummy_vrms_map, sharp_image)
        expected = np.mean(dummy_vrms_map) / np.mean(sharp_image) * np.ones(4) / 25.0
        npt.assert_allclose(vrms, expected, rtol=1e-3)



__author__ = "gipagano"


import numpy as np
import numpy.testing as npt
import pytest
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.nie_potential import NIE_POTENTIAL
from lenstronomy.LensModel.Profiles.spep import SPEP


class TestNIE_POTENTIAL(object):
    """Tests the NIE_POTENTIAL profile for different rotations."""

    def setup_method(self):
        self.nie_potential = NIE_POTENTIAL()
        self.spep = SPEP()

    def test_function(self):
        y = np.array([1.0, 2])
        x = np.array([0.0, 0.0])

        theta_E = 1.0
        theta_c = 0.0

        #############
        # no rotation
        #############

        e1, e2 = 0.05, 0.0
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the non-rotated output
        values = self.nie_potential.function(x, y, theta_E, theta_c, e1, e2)
        delta_pot = values[1] - values[0]
        values = self.spep.function(x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep)
        delta_pot_spep = values[1] - values[0]
        npt.assert_almost_equal(delta_pot, delta_pot_spep, decimal=4)

        ############
        # rotation 1
        ############

        e1, e2 = 0.05, 0.1
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        values = self.nie_potential.function(x, y, theta_E, theta_c, e1, e2)
        delta_pot = values[1] - values[0]
        values = self.spep.function(x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep)
        delta_pot_spep = values[1] - values[0]
        npt.assert_almost_equal(delta_pot, delta_pot_spep, decimal=4)

        ############
        # rotation 2
        ############

        e1, e2 = 0.15, 0.13
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        values = self.nie_potential.function(x, y, theta_E, theta_c, e1, e2)
        delta_pot = values[1] - values[0]
        values = self.spep.function(x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep)
        delta_pot_spep = values[1] - values[0]
        npt.assert_almost_equal(delta_pot, delta_pot_spep, decimal=4)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])

        theta_E = 1.0
        theta_c = 0.0

        #############
        # no rotation
        #############

        e1, e2 = 0.05, 0.0
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the non-rotated output
        f_x, f_y = self.nie_potential.derivatives(x, y, theta_E, theta_c, e1, e2)
        f_x_nie, f_y_nie = self.spep.derivatives(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_x, f_x_nie, decimal=4)
        npt.assert_almost_equal(f_y, f_y_nie, decimal=4)

        ############
        # rotation 1
        ############

        e1, e2 = 0.05, 0.1
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        f_x, f_y = self.nie_potential.derivatives(x, y, theta_E, theta_c, e1, e2)
        f_x_nie, f_y_nie = self.spep.derivatives(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_x, f_x_nie, decimal=4)
        npt.assert_almost_equal(f_y, f_y_nie, decimal=4)

        ############
        # rotation 2
        ############

        e1, e2 = 0.15, 0.13
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        f_x, f_y = self.nie_potential.derivatives(x, y, theta_E, theta_c, e1, e2)
        f_x_nie, f_y_nie = self.spep.derivatives(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_x, f_x_nie, decimal=4)
        npt.assert_almost_equal(f_y, f_y_nie, decimal=4)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])

        theta_E = 1.0
        theta_c = 0.0

        #############
        # no rotation
        #############

        e1, e2 = 0.05, 0.0
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the non-rotated output
        f_xx, f_xy, f_yx, f_yy = self.nie_potential.hessian(
            x, y, theta_E, theta_c, e1, e2
        )
        f_xx_nie, f_xy_nie, f_yx_nie, f_yy_nie = self.spep.hessian(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_xx, f_xx_nie, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_nie, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_nie, decimal=4)
        npt.assert_almost_equal(f_yx, f_yx_nie, decimal=4)

        ############
        # rotation 1
        ############

        e1, e2 = 0.05, 0.1
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        f_xx, f_xy, f_yx, f_yy = self.nie_potential.hessian(
            x, y, theta_E, theta_c, e1, e2
        )
        f_xx_nie, f_xy_nie, f_yx_nie, f_yy_nie = self.spep.hessian(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_xx, f_xx_nie, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_nie, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_nie, decimal=4)
        npt.assert_almost_equal(f_yx, f_yx_nie, decimal=4)

        ############
        # rotation 2
        ############

        e1, e2 = 0.15, 0.13
        eps = np.sqrt(e1**2 + e2**2)
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)

        # map the nie_potential input to the spep input
        gamma_spep = 2.0
        q_spep = np.sqrt(q)
        e1_spep, e2_spep = param_util.phi_q2_ellipticity(phi_G, q_spep)
        theta_E_conv = self.nie_potential._theta_q_convert(theta_E, q)
        theta_E_spep = theta_E_conv * np.sqrt(1 - eps) / ((1 - eps) / (1 + eps)) ** 0.25

        # compare the rotated output
        f_xx, f_xy, f_yx, f_yy = self.nie_potential.hessian(
            x, y, theta_E, theta_c, e1, e2
        )
        f_xx_nie, f_xy_nie, f_yx_nie, f_yy_nie = self.spep.hessian(
            x, y, theta_E_spep, gamma_spep, e1_spep, e2_spep
        )
        npt.assert_almost_equal(f_xx, f_xx_nie, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_nie, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_nie, decimal=4)
        npt.assert_almost_equal(f_yx, f_yx_nie, decimal=4)

    def test_static(self):
        x, y = 1.0, 1.0
        phi_G, q = 0.3, 0.8

        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = {"theta_E": 1.0, "theta_c": 0.1, "e1": e1, "e2": e2}
        f_ = self.nie_potential.function(x, y, **kwargs_lens)
        self.nie_potential.set_static(**kwargs_lens)
        f_static = self.nie_potential.function(x, y, **kwargs_lens)
        npt.assert_almost_equal(f_, f_static, decimal=8)
        self.nie_potential.set_dynamic()
        kwargs_lens = {"theta_E": 2.0, "theta_c": 0.1, "e1": e1, "e2": e2}
        f_dyn = self.nie_potential.function(x, y, **kwargs_lens)
        assert f_dyn != f_static


if __name__ == "__main__":
    pytest.main()



import astropy

if float(astropy.__version__[0]) < 5.0:
    from astropy.cosmology.core import vectorize_if_needed
else:
    Warning(
        "This routines are only supported for astropy version <5. Current version is %s."
        % astropy.__version__
    )
#
from scipy.integrate import quad


class CosmoInterp(object):
    """Class which interpolates the comoving transfer distance and then computes angular
    diameter distances from it This class is modifying the astropy.cosmology
    routines."""

    def __init__(self, cosmo):
        """

        :param cosmo: astropy.cosmology instance (version 4.0 as private functions need to be supported)
        """
        self._cosmo = cosmo

    def _integral_comoving_distance_z1z2(self, z1, z2):
        """Comoving line-of-sight distance in Mpc between objects at redshifts z1 and
        z2.

        The comoving distance along the line-of-sight between two
        objects remains constant with time for objects in the Hubble
        flow.

        Parameters
        ----------
        z1, z2 : array_like, shape (N,)
          Input redshifts.  Must be 1D or scalar.

        Returns
        -------
        d : `~astropy.units.Quantity`
          Comoving distance in Mpc between each input redshift.
        """

        f = lambda z1, z2: quad(
            self._cosmo._inv_efunc_scalar,
            z1,
            z2,
            args=self._cosmo._inv_efunc_scalar_args,
        )[0]
        return self._cosmo._hubble_distance * vectorize_if_needed(f, z1, z2)



__author__ = "aymgal, johannesulf"


from inspect import signature
from lenstronomy.Sampling.Samplers.base_nested_sampler import NestedSampler


__all__ = ["NautilusSampler"]


class NautilusSampler(NestedSampler):
    """Wrapper for the nautilus sampler by Johannes U. Lange.

    paper : https://arxiv.org/abs/2306.16923
    doc : https://nautilus-sampler.readthedocs.io
    """

    def __init__(
        self,
        likelihood_module,
        prior_type="uniform",
        prior_means=None,
        prior_sigmas=None,
        width_scale=1,
        sigma_scale=1,
        mpi=False,
        **kwargs
    ):
        """
        :param likelihood_module: likelihood_module like in likelihood.py (should be callable)
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to param cube
        :param prior_means: if prior_type is 'gaussian', mean for each param
        :param prior_sigmas: if prior_type is 'gaussian', std dev for each param
        :param width_scale: scale the widths of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by this factor
        :param mpi: Use MPI computing if `True`
        :param kwargs: kwargs directly passed to Sampler
        """
        self._check_install()
        super(NautilusSampler, self).__init__(
            likelihood_module,
            prior_type,
            prior_means,
            prior_sigmas,
            width_scale,
            sigma_scale,
        )

        if mpi:
            from schwimmbad import MPIPool
            import sys

            # use_dill=True not supported for some versions of schwimmbad
            pool = MPIPool(use_dill=True)
            if not pool.is_master():
                pool.wait()
                sys.exit(0)
            kwargs["pool"] = pool

        keys = [p.name for p in signature(self._nautilus.Sampler).parameters.values()]
        kwargs = {key: kwargs[key] for key in kwargs.keys() & keys}

        self._sampler = self._nautilus.Sampler(
            self.prior, self.log_likelihood, self.n_dims, **kwargs
        )

    def run(self, **kwargs):
        """Run the nautilus nested sampler.

        see https://nautilus-sampler.readthedocs.io for content of kwargs

        :param kwargs: kwargs directly passed to Sampler.run
        :return: samples, means, logZ, logZ_err, logL, results
        """
        print("prior type :", self.prior_type)
        print("parameter names :", self.param_names)

        keys = [p.name for p in signature(self._sampler.run).parameters.values()]
        kwargs = {key: kwargs[key] for key in kwargs.keys() & keys}
        self._sampler.run(**kwargs)
        points, log_w, log_l = self._sampler.posterior()
        log_z = self._sampler.evidence()

        return points, log_w, log_l, log_z

    def _check_install(self):
        try:
            import nautilus
        except ImportError:
            print(
                "Warning : nautilus not properly installed. \
                  You can get it with $pip install nautilus-sampler."
            )
        else:
            self._nautilus = nautilus



__author__ = "aymgal"

import pytest
import numpy as np

from lenstronomy.Sampling.Samplers.base_nested_sampler import NestedSampler


@pytest.fixture
def import_fixture(simple_einstein_ring_likelihood):
    """

    :param simple_einstein_ring_likelihood: fixture
    :return:
    """
    likelihood, kwargs_truths = simple_einstein_ring_likelihood
    prior_means = likelihood.param.kwargs2args(**kwargs_truths)
    prior_sigmas = np.ones_like(prior_means) * 0.1
    sampler = NestedSampler(likelihood, "gaussian", prior_means, prior_sigmas, 0.5, 0.5)
    return sampler, likelihood


class TestNestedSampler(object):
    """Test the fitting sequences."""

    def setup_method(self):
        pass

    def test_sampler(self, import_fixture):
        sampler, likelihood = import_fixture
        kwargs_run = {}
        try:
            sampler.run(kwargs_run)
        except Exception as e:
            assert isinstance(e, NotImplementedError)

    def test_sampler_init(self, import_fixture):
        _, likelihood = import_fixture
        sampler = NestedSampler(likelihood, "uniform", None, None, 1, 1)
        try:
            sampler = NestedSampler(
                likelihood, "gaussian", None, None, 1, 1
            )  # will raise an Error
        except Exception as e:
            assert isinstance(e, ValueError)
        try:
            sampler = NestedSampler(
                likelihood, "some_type", None, None, 1, 1
            )  # will raise an Error
        except Exception as e:
            assert isinstance(e, ValueError)

    def test_prior(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        cube = np.zeros(n_dims)
        try:
            sampler.prior(cube)
        except Exception as e:
            assert isinstance(e, NotImplementedError)

    def test_log_likelihood(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        args = np.nan * np.ones(n_dims)
        try:
            sampler.log_likelihood(args)
        except Exception as e:
            assert isinstance(e, NotImplementedError)


if __name__ == "__main__":
    pytest.main()
